;***********************************************************************************;
;***********************************************************************************;
;
; The almost completely commented BASIC4 ROM disassembly. v1.2
; By Simon Rowe <srowe@mose.org.uk>.


;***********************************************************************************;
;***********************************************************************************;
;
; BASIC zero page

VERCHK	= $0A			; load/verify flag
COUNT	= $0B
GARBFL	= $0F
LINNUM	= $14
TEMPST	= $19
INDEX	= $22
FRESPC	= $35			; string utility ptr
VARNAM	= $45			; current variable name
VARPNT	= $47			; current variable address
FORPNT	= $49
FAC1	= $61
FBUFPT	= $71
CHRGET	= $73			; increment and scan memory, BASIC byte get
CHRGOT	= $79			; scan memory, BASIC byte get


;***********************************************************************************;
;***********************************************************************************;
;
; KERNAL zero page

STATUS	= $90			; I/O status byte
VERCK	= $93			; load/verify flag
C3PO	= $94			; deferred character flag
BSOUR	= $95			; deferred character
XSAV	= $97
LDTND	= $98			; open file count
DFLTN	= $99			; input device number
DFLTO	= $9A			; output device number
PTR1	= $9E
TIME	= $A0			; jiffy clock
SAL	= $AC			; start pointer
EAL	= $AE			; end pointer
CMP0	= $B0
FNLEN	= $B7			; file name length
LA	= $B8			; logical file
SA	= $B9			; secondary address
FA	= $BA			; current device number
FNADR	= $BB			; file name pointer
STAL	= $C1
MEMUSS	= $C3
PNT	= $D1			; current screen line pointer
PNTR	= $D3			; cursor column
USER	= $F3			; colour RAM pointer low byte

BUF	= $0200			; input buffer
LAT	= $0259			; logical file table
FAT	= $0263			; device number table
SAT	= $026D			; secondary address table
KEYD	= $0277			; keyboard buffer
COLOR	= $0286			; current colour code
SHFLAG	= $028D			; keyboard shift/control flag
M51CDR	= $0294			; pseudo 6551 command register
IERROR	= $0300			; BASIC vector start


;***********************************************************************************;
;
; hardware equates

	; Standard VIA
VIA1PA1		= $9111		; VIA 1 DRA

	; Standard VIA (alternative address)
ALTVIA1PB	= $9350		; VIA 1 DRB
					; bit	function
					; ---	--------
					;  7	D8
					;  6	D7
					;  5	D6
					;  4	D5
					;  3	D4
					;  2	D3
					;  1	D2
					;  0	D1

ALTVIA1DDRB	= $9352		; VIA 1 DDRB
ALTVIA1T1CH	= $9355		; VIA 1 T1C_h
ALTVIA1ACR	= $935B		; VIA 1 ACR
					; bit	function
					; ---	--------
					;  7	T1 PB7 enabled/disabled
					;  6	T1 free run/one shot
					;  5	T2 clock PB6/ø2
					; 432	function
					; ---	--------
					; 000	shift register disabled
					; 001	shift in, rate controlled by T2
					; 010	shift in, rate controlled by ø2
					; 011	shift in, rate controlled by external clock
					; 100	shift out, rate controlled by T2, free run mode
					; 101	shift out, rate controlled by T2
					; 110	shift out, rate controlled by ø2
					; 111	shift out, rate controlled by external clock
					;  1	PB latch enabled/disabled
					;  0	PA latch enabled/disabled

ALTVIA1PCR	= $935C		; VIA 1 PCR
					; bit	function
					; ---	--------
					; 765	CB2 STROBE
					;  4	CB1 ACK

; The status bit is a not normal flag. It goes high if both an interrupt flag in the IFR
; and the corresponding enable bit in the IER are set. It can be cleared only by clearing
; all the active flags in the IFR or disabling all active interrupts in the IER.

ALTVIA1IFR	= $935D		; VIA 1 IFR
					; bit	function		cleared by
					; ---	--------		----------
					;  7	interrupt status	clearing all enabled interrupts
					;  6	T1 interrupt		read T1C_l, write T1C_h
					;  5	T2 interrupt		read T2C_l, write T2C_h
					;  4	CB1 transition		read or write port B
					;  3	CB2 transition		read or write port B
					;  2	8 shifts done		read or write the shift register
					;  1	CA1 transition		read or write port A
					;  0	CA2 transition		read or write port A


	; Cartridge PIA
PIAORA	= $9C00			; PIA PRA/DDRA
					; bit	function
					; ---	--------
					;  7	DIO8
					;  6	DIO7
					;  5	DIO6
					;  4	DIO5
					;  3	DIO4
					;  2	DIO3
					;  1	DIO2
					;  0	DIO1

PIACRA	= $9C01			; PIA CRA
					; bit	function
					; ---	--------
					;  7	CA1 transition
					;  6	CA2 transition
					;  5	CA2 direction, output
					;  4	CA2 pulse/level, level
					;  3	CA2 pin, ATN out
					;  2	ORA data direction or data
					;  1	CA1 high>low/low>high
					;  0	CA1 interrupt enable

PIAORB	= $9C02			; PIA PRB/DDRB
					; bit	function
					; ---	--------
					;  7	NDAC
					;  6	NRFD
					;  5	DAV
					;  4	EOI
					;  3
					;  2	??? SRQ in ???
					;  1
					;  0

PIACRB	= $9C03			; PIA CRB
					; bit	function
					; ---	--------
					;  7	CB1 transition
					;  6	CB2 transition
					;  5	CB2 direction
					;  4	CB2 high>low/low>high or pulse/level
					;  3	CB2 interrupt enable or ack/strobe
					;  2	ORB data direction or data
					;  1	CB1 high>low/low>high
					;  0	CB1 interrupt enable

	; BASIC entrypoints
WARMST	= $C002			; BASIC warm start entry point
ERROR	= $C437			; do error #.X then warm start (via vector)
ERROR2	= $C43A			; do error #.X then warm start
MAIN2	= $C483			; BASIC warm start
LNKPRG	= $C533			; rebuild BASIC line chaining
CRNCH2	= $C57C			; crunch BASIC tokens (via vector)
NEWSTT	= $C7AE			; interpreter inner loop
GOTO	= $C8A0			; perform GOTO
REM	= $C93B			; perform REM
FRMEVL	= $CD9E			; evaluate expression
FEVAL	= $CE86			; get arithmetic element
RPACHK	= $CEF7			; scan for ")"
LPACHK	= $CEFA			; scan for "("
COMCHK	= $CEFD			; scan for ","
SYNCHR	= $CEFF			; scan for CHR$(.A)
FNDVAR	= $D0E7			; either find or create variable
ALCSPAC	= $D4F4			; make space in string memory for string
DELST	= $D6A3			; evaluate string
MAKADR	= $D7F7			; convert FAC1 to integer in temporary integer
INITBA	= $E3A4			; initialize BASIC RAM locations
FREMSG	= $E404			; print start up message
PATCH1	= $E4BC			; get secondary address and print "SEARCHING..."
PATCH2	= $E4C1			; set LOAD address if secondary address = 0
	
	; KERNAL entrypoints
INITSK	= $E518			; initialize hardware
SCRNOUT	= $E742			; output character
PUTSCRN	= $EAAA			; save character and colour to screen @ cursor
COLORSYN = $EAB2		; calculate pointer to colour RAM
IRQ	= $EABF			; IRQ handler
FTALK	= $EE14			; command a serial bus device to TALK
FLISTEN	= $EE17			; command devices on the serial bus to LISTEN
SRSEND	= $EE49			; Tx byte on serial bus
FSECOND	= $EEC0			; send secondary address after LISTEN
SCATN	= $EEC5			; set serial ATN high
FTKSA	= $EECE			; send secondary address after TALK
FUNTLK	= $EEF6			; command the serial bus to UNTALK
FUNLSN	= $EF04			; command the serial bus to UNLISTEN
FACPTR	= $EF19			; input a byte from the serial bus
RSOUTSAV = $F0ED		; send byte to RS-232 buffer
RSNXTIN	= $F14F			; get byte from RS-232 buffer
FGETIN	= $F1F5			; get a character from the input device
FCHRIN	= $F20E			; input character from channel
FCHROUT	= $F27A			; output a character to channel
FNDFLNO	= $F3CF			; find file
SETFLCH	= $F3DF			; set file details from table,.X
FCLALL	= $F3EF			; close all channels and files
FCLRCHN	= $F3F3			; close input and output channels
SERNAME	= $F495			; send secondary address and filename
FSAVE2	= $F685			; save
FUDTIM	= $F734			; increment real time clock
FSTOP	= $F770			; scan stop key
FE_2MNYF = $F77E		; display too many files
FE_ALOPN = $F781		; display file already open
FE_NTOPN = $F784		; display file not open
FE_NTFND = $F787		; display file not found
FE_DVNTP = $F78A		; display device not present
FE_NTINP = $F78D		; display not input file
FE_MISFN = $F793		; display missing file name
RD300	= $FBD2			; copy I/O start address to buffer address
VPRTY	= $FD11			; check read/write pointer
WRT62	= $FD1B			; increment read/write pointer
INITMEM	= $FD8D			; initialize and test RAM
INITVIA	= $FDF9			; initialize I/O registers
ORIOST	= $FE6A			; OR into I/O status byte
NMI2	= $FEAD			; NMI handler
BREAK	= $FED2			; BRK handler
SCNKEY	= $FF9F			; scan the keyboard
CLOSE	= $FFC3			; close a specified logical file (via vector)
CLRCHN	= $FFCC			; close input and output channels (via vector)
CHRIN	= $FFCF			; input character from channel (via vector)
CHROUT	= $FFD2			; output character to channel (via vector)
LOAD	= $FFD5			; load RAM from a device (via vector)
STOP	= $FFE1			; scan stop key (via vector)


;***********************************************************************************;
;
; BASIC keyword token values

TK_GOTO		= $89			; GOTO token
TK_IF		= $8B			; IF token
TK_ON		= $91			; ON token
TK_TO		= $A4			; TO token
TK_THEN		= $A7			; THEN token

TK_PI		= $FF			; PI token


;***********************************************************************************;
;
; error messages

ER_STOP		= $00
ER_FNOTOPEN	= $03
ER_DEVNOTP	= $05
ER_ILLQUAN	= $0E
ER_STR2LONG	= $17


;***********************************************************************************;
;
; floating point accumulator offsets

FAC_EXPT	= $00
FAC_MANT	= $01
FAC_SIGN	= $05


;***********************************************************************************;
;
; private variables

SAVEREG		= $02			; MONITOR register save area
PARSEST		= $C1			; disk command parameter parse state
DRIVENO2	= $C2			; secondary disk number
DRIVENO1	= $C3			; primary disk number
BUSCONF		= $0334			; serial/IEEE bus selection
STVALID		= $0335			; disk status variables valid
MONERRV		= $0336			; MONITOR error vector
RECLEN		= $033D			; record length/disk identifier flag
DISKID		= $033F			; disk identifier (2 bytes)
TEMPLEN		= $0341			; temporary length
TEMPSTR		= $0342			; temporary string
CMDBUF		= $0353			; disk command buffer


;***********************************************************************************;
;***********************************************************************************;
;
; BASIC4 ROM start

	* = $A000

	.WORD B4COLD
	.WORD B4WARM
	.BYTE "A0",'C'+$80,'B'+$80,'M'+$80


;***********************************************************************************;
;
; BASIC4 warm start entry point

B4COLD
	JSR INITMEM		; initialize and test RAM
	JSR B4VECTS		; initialize BASIC and KERNAL vectors
	STY STVALID		; set disk status valid
	LDA #$03		; set IEEE bus
	STA BUSCONF		; set bus config
	JSR INITVIA		; initialize I/O registers
	JSR INITSK		; initialize hardware
	CLI			; enable interrupts
	JSR INITBA		; initialize BASIC RAM locations
	JSR FREMSG		; print start up message and initialize memory pointers
	LDY #$08		; set base address and index to 16 / 2
	STY PNT			; set current screen line pointer low byte
	LDA #'4'		; screen code for "4"
	STA (PNT),Y		; update banner to read "BASIC V4"
	LDA #<MONERR		; get MONITOR error routine low byte
	STA MONERRV		; set MONITOR error vector low byte
	LDA #>MONERR		; get MONITOR error routine high byte
	STA MONERRV+1		; set MONITOR error vector high byte
	JSR SCNKEY		; scan keyboard
	LDA SHFLAG		; get keyboard shift/control/C= flag
	AND #$02		; mask [C=] bit
	BEQ LA043		; branch if not [C=]
	INC BUSCONF		; set bus config to serial
LA043	JSR INITIEVIA		; initialize cartridge I/O registers
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	JMP $E381		; continue normal BASIC coldstart


;***********************************************************************************;
;
; initialize BASIC and KERNAL vectors

B4VECTS
	LDY #$33		; set byte count
LA051	LDA ALLVECTS,Y		; get byte from table
	STA IERROR,Y		; save byte to RAM
	DEY			; decrement index
	BPL LA051		; loop if more to do
	RTS


;***********************************************************************************;
;
; BASIC vectors, these are copied to RAM from $0300 onwards

ALLVECTS
	.WORD ERROR2		; BASIC vector - print error message
	.WORD MAIN2		; BASIC vector - main command processor
	.WORD B4CRNCH		; BASIC vector - tokenise keywords
	.WORD B4QPLOP		; BASIC vector - list program
	.WORD B4GONE		; BASIC vector - execute next command
	.WORD FEVAL		; BASIC vector - get value from line
	
DFLTNAME
	.BYTE "0:*",$00		; default load file name
RUNCR	.BYTE "RUN",$0D		; second part of RUNTB


;***********************************************************************************;
;
; KERNAL vectors

	.WORD IRQ		; IRQ vector
	.WORD B4BREAK		; BRK vector
	.WORD NMI2		; NMI vector

	.WORD B4OPEN		; KERNAL vector - open a logical file
	.WORD B4CLOSE		; KERNAL vector - close a specified logical file
	.WORD B4CHKIN		; KERNAL vector - open channel for input
	.WORD B4CHKOUT		; KERNAL vector - open channel for output
	.WORD B4CLRCHN		; KERNAL vector - close input and output channels
	.WORD B4CHRIN		; KERNAL vector - input character from channel
	.WORD B4CHROUT		; KERNAL vector - output character to channel
	.WORD FSTOP		; KERNAL vector - scan stop key
	.WORD B4GETIN		; KERNAL vector - get character from keyboard queue
	.WORD FCLALL		; KERNAL vector - close all channels and files
	.WORD BREAK		; User vector

	.WORD B4LOAD		; KERNAL vector - load
	.WORD B4SAVE		; KERNAL vector - save


;***********************************************************************************;
;
; BASIC4 warm start entry point

B4WARM
	BIT VIA1PA1		; test VIA 1 DRA
	JSR FUDTIM		; increment the real time clock
	JSR STOP		; scan stop key
	BNE LA0A0		; if no stop then exit
	JSR B4VECTS		; initialize BASIC and KERNAL vectors
	JMP $FED5		; jump to normal warm start
LA0A0	JMP $FEFF		; restore registers and exit interrupt


;***********************************************************************************;
;
; get file details

GETLFN
	LDA #$00		; clear .A
	STA STATUS		; clear I/O status byte
	JSR FNDFLNO		; find file
	BEQ LA0B1		; branch if file opened
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
	JMP FE_NTOPN		; do file not open error and return
LA0B1	JSR SETFLCH		; set file details from table,X
LA0B4	LDA FA			; get device number
	CMP #$04		; compare device with first serial device
	RTS


;***********************************************************************************;
;
; command a device on either the serial or IEEE bus to TALK

DBTALK
	JSR QUERYBUS		; query bus to use
	LDA FA			; get device number
	BCS LA0D3		; branch if IEEE
	STA STVALID		; reset disk status valid
	JMP FTALK		; command a serial bus device to TALK


;***********************************************************************************;
;
; command devices on either the serial or IEEE bus to LISTEN

DBLISTEN
	JSR QUERYBUS		; query bus to use
	LDA FA			; get device number
	BCS LA0D6		; branch if IEEE
	STA STVALID		; reset disk status valid
	JMP FLISTEN		; command devices on the serial bus to LISTEN

LA0D3	JMP IETALK		; command an IEEE bus device to TALK
LA0D6	JMP IELISTEN		; command devices on the IEEE bus to LISTEN
	; dead code
	LDA #$B8		; CB2 output, CB2 level, CB2 high, DDR, CB1 high>low
	STA PIACRB		; set PIA CRB
	LDA #$0B		; IIII OIOO, NDAC in, NRFD in, DAV in, EOI in, SRQ in
	STA PIAORB		; set PIA DDRB
	LDA #$BC		; CB2 output, CB2 level, CB2 high, data register, CB1 high>low
	STA PIACRB		; set PIA CRB
LA0E8	LDA PIAORB		; get PIA PRB
	AND #$04		; mask SRQ
	BNE LA0FA		; if SRQ high then send control byte
	LDA TIME+2		; get current jiffies
	ADC #$3C		; add 60 jiffies
LA0F3	CMP TIME+2		; compare with current jiffies
	BNE LA0F3		; loop until timeout
	BEQ LA0E8		; repeat while SRQ low, branch always


;***********************************************************************************;
;
; send control byte

LA0F9	PHA			; save command
LA0FA	LDA #$FB		; xxxx 1x11, ???
	STA PIAORB		; set PIA PRB
	LDA #$B8		; Cx2 output, Cx2 level, Cx2 high, DDR, Cx1 high>low
	STA PIACRA		; set PIA CRA
	STA PIACRB		; set PIA CRB
	LDA #$DF		; OOIO OOOO, NDAC out, NRFD out, DAV in, EOI out, SRQ out
	STA PIAORB		; set PIA DDRB
	LDA #$FF		; all outputs
	STA PIAORA		; set PIA DDRA
	LDA #$BC		; Cx2 output, Cx2 level, Cx2 high, data register, Cx1 high>low
	STA PIACRA		; set PIA CRA
	STA PIACRB		; set PIA CRB
	LDA C3PO		; get deferred character flag
	BPL LA127		; branch if no deferred character
	LDA #$EB		; NDAC high, NRFD high, EOI low, SRQ low
	STA PIAORB		; set PIA PRB
	JSR DBSEND		; Tx byte on either serial or IEEE bus
	LSR C3PO		; clear deferred character flag
LA127	LDA #$00		; T1 disabled, T2 timed, shift register disabled, PA & PB latch disabled
	STA ALTVIA1ACR		; set VIA 1 ACR
LA12C	LDA PIAORB		; get PIA PRB
	AND #$20		; mask DAV
	BEQ LA12C		; loop until DAV high
	LDA #$B4		; CA2 output, CA2 level, CA2 low, data register, CA1 high>low
	STA PIACRA		; set PIA CRA, ATN low
	PLA			; restore command
	ORA FA			; OR with device number
	STA BSOUR		; save byte then ..


;***********************************************************************************;
;
; Tx byte on either serial or IEEE bus

DBSEND
	JSR QUERYBUS		; query bus to use
	BCS LA145		; branch if IEEE
	JMP SRSEND		; Tx byte on serial bus

LA145	JMP LB2E5		; output byte on IEEE bus
	NOP
	NOP
	LDA #$3F		; IIOO OOOO, NDAC in, NRFD in, DAV out, EOI out, SRQ out
	STA PIAORB		; set PIA DDRB
	STA STVALID		; reset disk status valid
	LDA #$BC		; CB2 output, CB2 level, CB2 high, data register, CB1 high>low
	STA PIACRB		; set PIA CRB
	BIT PIAORB		; test PIA PRB
	BPL LA15E		; if NDAC low continue
	BVS LA189		; else if NRFD high do device not present
LA15E	BIT PIAORB		; test PIA PRB
	BVC LA15E		; loop until NRFD goes low
	LDA #$FF		; set for 65,280 cycles
	STA ALTVIA1T1CH		; set VIA 1 T1C_h
	EOR BSOUR		; invert output byte
	STA PIAORA		; set PIA PRA, DIO1-8 out
	LDA PIAORB		; get PIA PRB
	AND #$1F		; set DAV low
	STA PIAORB		; set PIA PRB
LA175	BIT ALTVIA1IFR		; test VIA 1 IFR
	BVS LA18C		; if T1 timeout do time out write
	BIT PIAORB		; test PIA PRB
	BPL LA175		; loop while NDAC low
	JSR IEDAVH		; set IEEE DAV high
LA182	LDA #$FF		; set data to $00
	STA PIAORA		; set PIA PRA
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; device not present

LA189	LDA #$80		; set device not present bit
	.BYTE $2C		; makes next line BIT $01A9


;***********************************************************************************;
;
; timeout on serial bus

LA18C	LDA #$01		; set time out write bit
	JSR ORIOST		; OR into I/O status byte
	JSR DBSCATN		; set serial or IEEE ATN high
	BNE LA182		; flag ok and exit, branch always


;***********************************************************************************;
;
; output a byte to either the serial or IEEE bus

DBCIOUT
	PHA			; save byte to output
LA197	LDA C3PO		; get deferred character flag
	BPL LA19E		; branch if no deferred character
	JSR DBSEND		; Tx byte on either serial or IEEE bus
LA19E	SEC			; set carry
	ROR C3PO		; shift into deferred character flag
LA1A1	PLA			; restore byte
	STA BSOUR		; save deferred byte
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; output a character to channel

; This routine will output a character to an already opened channel. Use the OPEN
; routine, OPEN, and the CHKOUT routine, to set up the output channel before calling
; this routine. If these calls are omitted, data will be sent to the default output
; device, device 3, the screen. The data byte to be output is loaded into the accumulator,
; and this routine is called. The data is then sent to the specified output device.
; The channel is left open after the call.

B4CHROUT
	PHA			; save the character to send
	LDA DFLTO		; get output device number
	CMP #$04		; compare device number with first external device
	BCC LA1CC		; branch if internal device
	BNE LA197		; branch if not device 4, printer
	JSR QUERYBUS		; query bus to use
	BVC LA197		; branch if not user port
	LDA #$AE		; CB2 pulse, CB1 -ve edge, CA2 high, CA1 -ve edge
	STA ALTVIA1PCR		; set VIA 1 PCR
	LDA #$FF		; all outputs
	STA ALTVIA1DDRB		; set VIA 1 DDRB
	PLA			; restore saved character
	STA ALTVIA1PB		; set VIA 1 DRB
	PHA			; save the character to send
LA1C3	LDA ALTVIA1IFR		; get VIA1 IFR
	AND #$10		; mask CB1 interrupt
	BEQ LA1C3		; loop until ACK high
	BNE LA1A1		; restore .A and exit, branch always

LA1CC	PLA			; restore character to send
	JMP FCHROUT		; output using KERNAL routine and return

LA1D0	LDA #$0D		; set [CR]
LA1D2	PHA			; save character
	JSR RESTOUT		; restore previous output device
	PLA			; restore character
	JSR CHROUT		; output character using vectored routine


;***********************************************************************************;
;
; close input and output channels

; This routine is called to clear all open channels and restore the I/O channels to
; their original default values. It is usually called after opening other I/O
; channels and using them for input/output operations. The default input device is
; 0, the keyboard. The default output device is 3, the screen.

B4CLRCHN
	JSR QUERYBUS		; query bus to use
	BCC LA200		; branch if serial
	LDA DFLTO		; get output device number
	CMP #$04		; compare with first external device
	BCC LA1F3		; branch if internal device
	BNE LA1EC		; branch if not device 4, printer
	JSR QUERYBUS		; query bus to use
	BVS LA1F3		; branch if user port
LA1EC	JSR DBUNLSN		; command the serial or IEEE bus to UNLISTEN
	LDA #$03		; set screen
	STA DFLTO		; set output device number
LA1F3	LDA DFLTN		; get input device number
	CMP #$04		; compare with first external device
	BCC LA200		; branch if internal device
	JSR DBUNTLK		; command the serial or IEEE bus to UNTALK
	LDA #$00		; set keyboard
	STA DFLTN		; set input device number
LA200	JMP FCLRCHN		; close input and output channels


;***********************************************************************************;
;
; open a logical file

; This routine is used to open a logical file. Once the logical file is set up it
; can be used for input/output operations. Most of the I/O KERNAL routines call on
; this routine to create the logical files to operate on. No arguments need to be
; set up to use this routine, but both the SETLFS and SETNAM KERNAL routines must
; be called before using this routine.

B4OPEN
	LDX LA			; get logical file
	BNE LA20A		; branch if there is a file
	JMP FE_NTINP		; else do not input file error and return
LA20A	JSR FNDFLNO		; find file
	BNE LA212		; branch if file not found
	JMP FE_ALOPN		; else do file already open error and return
LA212	LDX LDTND		; get open file count
	CPX #$0A		; compare with max
	BCC LA21B		; branch if less
	JMP FE_2MNYF		; else do too many files error and return
LA21B	INC LDTND		; increment open file count
	LDA LA			; get logical file
	STA LAT,X		; save to logical file table
	LDA SA			; get secondary address
	ORA #$60		; OR with the OPEN CHANNEL command
	STA SA			; set secondary address
	STA SAT,X		; save to secondary address table
	LDA FA			; get device number
	STA FAT,X		; save to device number table
	BEQ LA28B		; do ok exit if keyboard
	CMP #$03		; compare device number with screen
	BEQ LA28B		; do ok exit if screen
	BCS LA23B		; branch if bus device
	JMP $F444		; jump to KERNAL to handle tape or RS-232
LA23B	JSR DBSERNAME		; send secondary address and filename
	RTS


;***********************************************************************************;
;
; open a channel for input

; Any logical file that has already been opened by the OPEN routine can be defined as
; an input channel by this routine. The device on the channel must be an input device
; or an error will occur and the routine will abort.

B4CHKIN
	JSR GETLFN		; get file details
	BCS LA247		; branch if external device
	JMP $F2D2		; jump to KERNAL to handle internal devices
LA247	TAX			; copy device address
	JSR DBTALK		; command a device on either the serial or IEEE bus to TALK
	BIT STATUS		; test I/O status byte
	BMI LA2B3		; branch if device not present
	LDA SA			; get secondary address
	BMI LA256		; branch if no address
	JSR DBTKSA		; send secondary address after TALK
LA256	JSR DBSCATN		; set serial or IEEE ATN high
	TXA			; copy device address
	BIT STATUS		; test I/O status byte
	BMI LA2B3		; branch if device not present
	STA DFLTN		; set input device number
	CLC			; flag ok
LA261	RTS


;***********************************************************************************;
;
; restore previous output device

RESTOUT
	LDA EAL			; get saved output device
	CMP #$03		; compare with screen
	BEQ LA261		; if screen then exit
	LDX EAL+1		; get saved logical file then ..


;***********************************************************************************;
;
; open a channel for output

; Any logical file that has already been opened by the OPEN routine can be defined
; as an output channel by this routine the device on the channel must be an output
; output device or an error will occur and the routine will abort.

B4CHKOUT
	JSR GETLFN		; get file details
	BCC LA28D		; branch if internal device
	BNE LA276		; branch if not device 4, printer
	JSR QUERYBUS		; query bus to use
	BVS LA289		; branch if user port
LA276	TAX			; copy device number
	JSR DBLISTEN		; command devices on either the serial or IEEE bus to LISTEN
	LDA SA			; get secondary address
	BMI LA281		; branch if address to send
	JSR DBSECOND		; send secondary address after LISTEN
LA281	JSR LA35C		; set either serial or IEEE bus ATN high
	TXA			; copy device number
	BIT STATUS		; test I/O status byte
	BMI LA2B3		; branch if device not present
LA289	STA DFLTO		; save output device number
LA28B	CLC			; flag ok
	RTS

LA28D	JMP $F314		; jump to KERNAL to handle internal devices


;***********************************************************************************;
;
; send secondary address and filename

DBSERNAME
	JSR QUERYBUS		; query bus to use
	BCS LA29B		; branch if IEEE
	INC STVALID		; reset disk status valid
	JMP SERNAME		; send secondary address and filename
LA29B	JMP IESERNAME
	NOP
	LDY FNLEN		; get file name length
	BEQ LA28B		; ok exit if null
	JSR DBLISTEN		; command devices on either the serial or IEEE bus to LISTEN
	LDA SA			; get the secondary address
	ORA #$F0		; OR with the OPEN command
	JSR DBCMDATN		; send command byte then set ATN high
	LDA STATUS		; get I/O status byte
	BPL LA2B6		; branch if device present
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
LA2B3	JMP FE_DVNTP		; do device not present error and return
LA2B6	LDY #$00		; clear index
LA2B8	LDA (FNADR),Y		; get file name byte
	JSR DBCIOUT		; output a byte to either the serial or IEEE bus
	INY			; increment index
	CPY FNLEN		; compare with file name length
	BNE LA2B8		; loop if not all done
	BEQ DBUNLSN		; else command the serial or IEEE bus to UNLISTEN


;***********************************************************************************;
;
; close a specified logical file

; This routine is used to close a logical file after all I/O operations have been
; completed on that file. This routine is called after the accumulator is loaded
; with the logical file number to be closed, the same number used when the file was
; opened using the OPEN routine.

B4CLOSE
	JSR $F3D4		; find file .A
	BNE LA28B		; if not found the file was closed so just flag ok
	JSR SETFLCH		; set file details from table,.X
	TXA			; copy file index to .A
	PHA			; save file index
	JSR LA0B4		; compare device number with first serial device
	BCS LA2D6		; external device so continue
	JMP $F356		; else jump to KERNAL to close
LA2D6	JSR LA2DF		; close the serial or IEEE bus device
	JMP $F3B1		; restore file index and close file

LA2DC	JMP $F6DA		; close the serial bus device

; close the serial or IEEE bus device

LA2DF	JSR QUERYBUS		; query bus to use
	BCC LA2DC		; if serial close the serial bus device
	BIT SA			; test the secondary address
	BMI LA28B		; if already closed just exit
	JSR DBLISTEN		; command devices on either the serial or IEEE bus to LISTEN
	LDA SA			; get secondary address
	AND #$EF		; mask the channel number
	ORA #$E0		; OR with the CLOSE command
	JSR DBCMDATN		; send command byte then set ATN high


;***********************************************************************************;
;
; command the serial or IEEE bus to UNLISTEN

DBUNLSN
	JSR QUERYBUS		; query bus to use
	BCS LA30C		; branch if IEEE
	JMP FUNLSN		; command the serial bus to UNLISTEN


;***********************************************************************************;
;
; command the serial or IEEE bus to UNTALK

DBUNTLK
	JSR QUERYBUS		; query bus to use
	BCS LA304		; branch if IEEE
	JMP FUNTLK		; command the serial bus to UNTALK

; command the IEEE bus to UNTALK

LA304	JMP IEUNTLK		; send UNTALK command on IEEE bus
	NOP
	NOP
	LDA #$5F		; set the UNTALK command
	.BYTE $2C		; makes next line BIT $3FA9

; command the IEEE bus to UNLISTEN

LA30C	LDA C3PO		; get deferred character flag
	BPL DOUNLSN		; branch if no deferred character
	JSR LB322		; output final byte on IEEE bus
	LSR C3PO		; clear deferred character flag
DOUNLSN JMP IEUNLSN
	LDA #$0B		; IIII OIOO, NDAC in, NRFD in, DAV in, EOI in, SRQ in
	STA PIAORB		; set PIA DDRB
LA31B	JMP LB398		; set IEEE ATN high
	NOP
	NOP
	RTS


;***********************************************************************************;
;
; set serial or IEEE ATN high

DBSCATN
	JSR QUERYBUS		; query bus to use
	BCC LA361		; if serial ..
	JMP LB398		; set IEEE ATN high
	BNE LA31B		; set IEEE ATN high, branch always


;***********************************************************************************;
;
; set IEEE DAV high

IEDAVH
	LDA #$3B		; DAV high, EOI high
	STA PIAORB		; set PIA PRB
	LDA #$B8		; CB2 output, CB2 level, CB2 high, DDR, CB1 high>low
	STA PIACRB		; set PIA CRB
	LDA #$FF		; all outputs
	STA PIAORB		; set PIA DDRB
	LDA #$BC		; CB2 output, CB2 level, CB2 high, data register, CB1 high>low
	STA PIACRB		; set PIA CRB
	RTS


;***********************************************************************************;
;
; send secondary address after TALK

DBTKSA
	STA SA			; set secondary address
	JSR QUERYBUS		; query bus to use
	BCS TKPATCH		; branch if IEEE
	JMP FTKSA		; send secondary address after TALK
TKPATCH
	JMP IETKSA

;***********************************************************************************;
;
; send secondary address after LISTEN
	
DBSECOND
	STA SA			; set secondary address
	JSR QUERYBUS		; query bus to use
	BCS DBCMDATN		; if IEEE then send command byte and set ATN high
	JMP FSECOND		; send secondary address after LISTEN


;***********************************************************************************;
;
; send command byte then set ATN high

DBCMDATN
	STA BSOUR		; save deferred byte
	JSR DBSEND		; Tx byte on either serial or IEEE bus
LA35C	JSR QUERYBUS		; query bus to use
	BCS LA31B		; if IEEE set IEEE ATN high
LA361	JMP SCATN		; else set serial ATN high

LA364	JMP PATCH_NTFND		; do file not found error and return


;***********************************************************************************;
;
; load RAM from a device

; This routine will load data bytes from any input device directly into the memory
; of the computer. It can also be used for a verify operation comparing data from a
; device with the data already in memory, leaving the data stored in RAM unchanged.

B4LOAD
	STA VERCK		; save load/verify flag
	LDY #$03		; set index to string length
LA36B	LDA KEYD,Y		; get character from keyboard buffer
	CMP RUNCR,Y		; compare with "RUN"
	BNE LA38A		; branch if no match
	DEY			; decrement index
	BPL LA36B		; loop until done
	LDA FNLEN		; get file name length
	BNE LA38A		; branch if not null name
	LDA #$08		; set device to 8, first disk
	STA FA			; set device number
	LDA #$03		; set default file name length
	STA FNLEN		; set file name length
	LDA #>DFLTNAME		; get default name high byte
	STA FNADR+1		; set file name pointer high byte
	LDA #<DFLTNAME		; get default name low byte
	STA FNADR		; set file name pointer low byte
LA38A	LDA #$00		; clear .A
	STA STATUS		; clear I/O status byte
	JSR LA0B4		; compare device number with first serial device
	BCS LA396		; external device so continue
	JMP $F54F		; else jump to KERNAL load
LA396	LDY FNLEN		; get file name length
	BNE LA39D		; branch if not null name
	JMP FE_MISFN		; else do missing file name error and return
LA39D	JSR PATCH1		; get secondary address and print "Searching..."
	LDA #$60		; set secondary address to $00
	STA SA			; save secondary address
	JSR DBSERNAME		; send secondary address and filename
	JSR DBTALK		; command a device on either the serial or IEEE bus to TALK
	LDA SA			; get secondary address
	JSR DBTKSA		; send secondary address after TALK
	JSR DBACPTR		; input a byte from either the serial or IEEE bus
	STA EAL			; save program start address low byte
	LDA STATUS		; get I/O status byte
	LSR			; shift time out read ..
	LSR			; .. into carry bit
	BCS LA364		; if timed out go do file not found error and return
	JSR DBACPTR		; input a byte from either the serial or IEEE bus
	STA EAL+1		; save program start address high byte
	JSR PATCH2		; set LOAD address if secondary address = 0
LA3C2	LDA #$FD		; mask xxxx xx0x, clear time out read bit
	AND STATUS		; mask I/O status byte
	STA STATUS		; set I/O status byte
	JSR STOP		; scan stop key, return Zb = 1 = [STOP]
	BNE LA3D3		; branch if not [STOP]
	JSR LA3FF		; else close the serial or IEEE bus device
	JMP $F6CE		; flag terminated by STOP and exit
LA3D3	JSR DBACPTR		; input a byte from either the serial or IEEE bus
	TAX			; copy byte
	LDA STATUS		; get I/O status byte
	AND #$02		; mask status with time out read
	BNE LA3C2		; if timed out clear I/O status and retry
	TAY			; clear .Y
	TXA			; copy received byte back
	LDX VERCK		; get load/verify flag
	BEQ LA3ED		; branch if load
	CMP (EAL),Y		; compare byte with previously loaded byte
	BEQ LA3EF		; branch if match
	LDA #$10		; else set read error bit
	JSR ORIOST		; OR into I/O status byte
	.BYTE $2C		; makes next line BIT $AE91
LA3ED	STA (EAL),Y		; save byte to memory
LA3EF	INC EAL			; increment save pointer low byte
	BNE LA3F5		; if no rollover skip the high byte increment
	INC EAL+1		; else increment save pointer high byte
LA3F5	BIT STATUS		; test I/O status byte
	BVC LA3C2		; loop if not end of file
	JSR LA3FF		; command the serial or IEEE bus to UNTALK then close
	JMP $F641		; flag ok and get the LOAD end pointer

LA3FF	JSR DBUNTLK		; command the serial or IEEE bus to UNTALK
	JSR LA2DF		; close the serial or IEEE bus device
	JMP LNKPRG		; rebuild BASIC line chaining and return

LA408	LDA #$02		; set time out read bit
	JSR ORIOST		; OR into I/O status byte
	JSR DBSCATN		; set serial or IEEE ATN high
LA410	LDA #$0D		; return [EOL]
	RTS

LA413	JMP FGETIN		; get character from keyboard queue
LA416	JMP FCHRIN		; input character from channel


;***********************************************************************************;
;
; get a character from the input device

; In practice this routine operates identically to the CHRIN routine for all devices
; except for the keyboard. If the keyboard is the current input device this routine
; will get one character from the keyboard buffer. It depends on the IRQ routine to
; read the keyboard and put characters into the buffer.

B4GETIN
	LDA DFLTN		; get input device number
	CMP #$04		; compare with first external device
	BCC LA413		; branch if internal device


;***********************************************************************************;
;
; input character from channel

; This routine will get a byte of data from the channel already set up as the input
; channel by the CHKIN routine.

; If CHKIN has not been used to define another input channel the data is expected to be
; from the keyboard. the data byte is returned in the accumulator. The channel remains
; open after the call.

B4CHRIN
	LDA #$03		; set .A to screen
	CMP DFLTN		; compare input device number with screen
	BCS LA416		; branch if >= screen
	LDA STATUS		; get I/O status byte
	BNE LA410		; branch if error


;***********************************************************************************;
;
; input a byte from either the serial or IEEE bus

DBACPTR
	JSR QUERYBUS		; query bus to use
	BCS LA431		; branch if IEEE
	JMP FACPTR		; input a byte from the serial bus and return

LA431	JSR IEACPTR		; input a byte from the IEEE bus
	LDA BSOUR		; get byte received
	RTS
	LDA #$CF		; OOII OOOO, NDAC out, NRFD out, DAV in, EOI in, ATN out
	STA PIAORB		; set PIA DDRB
	LDA #$00		; all inputs
	STA PIAORA		; set PIA DDRA
	LDA #$3C		; Cx2 output, Cx2 level, Cx2 high, data register, Cx1 high>low
	STA PIACRB		; set PIA CRB
	STA PIACRA		; set PIA CRA
	LDA #$4B		; NDAC low, NRFD high
	STA PIAORB		; set PIA PRB
	LDA #$FF		; set for 65,280 cycles
	STA ALTVIA1T1CH		; set VIA 1 T1C_h
LA455	BIT ALTVIA1IFR		; test VIA 1 IFR
	BVS LA408		; if T1 timeout then set time out read and exit
	LDA PIAORB		; get PIA PRB
	AND #$20		; mask DAV
	BNE LA455		; loop if DAV high
	LDA #$0B		; NRFD low
	STA PIAORB		; set PIA PRB
	LDA PIAORB		; get PIA PRB
	AND #$10		; mask EOI
	BNE LA472		; branch if not EOI
	LDA #$40		; set EOI bit
	JSR ORIOST		; OR into I/O status byte
LA472	LDA PIAORA		; get PIA PRA
	EOR #$FF		; invert it
	PHA			; save byte received
	LDA #$8B		; NDAC high, NRFD low
	STA PIAORB		; set PIA PRB
LA47D	LDA PIAORB		; get PIA PRB
	AND #$20		; mask DAV
	BEQ LA47D		; loop if DAV low
	LDA #$2B		; NDAC low
	STA PIAORB		; set PIA PRB
	PLA			; restore byte received
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; save RAM to device, .A = index to start address, .X.Y = end address low/high

; This routine saves a section of memory. Memory is saved from an indirect address
; on page 0 specified by A, to the address stored in .X.Y, to a logical file. The
; SETLFS and SETNAM routines must be used before calling this routine. However, a
; file name is not required to SAVE to device 1, the cassette. Any attempt to save to
; other devices without using a file name results in an error.

B4SAVE
	JSR LA0B4		; compare device number with first serial device
	BCS LA494		; external device so continue
	JMP FSAVE2		; else jump to KERNAL save routine
LA494	LDA #$61		; set secondary address to $01
	STA SA			; save secondary address
	LDY FNLEN		; get file name length
	BNE LA49F		; branch if filename not null
	JMP FE_MISFN		; else do missing file name error and return
LA49F	JSR DBSERNAME		; send secondary address and filename
	JSR DBLISTEN		; command devices on either the serial or IEEE bus to LISTEN
	LDA SA			; get secondary address
	JSR DBSECOND		; send secondary address after LISTEN
	LDY #$00		; clear index
	JSR RD300		; copy I/O start address to buffer address
	LDA SAL			; get buffer address low byte
	JSR DBCIOUT		; output a byte to either the serial or IEEE bus
	LDA SAL+1		; get buffer address high byte
	JSR DBCIOUT		; output a byte to either the serial or IEEE bus
LA4B9	JSR VPRTY		; check read/write pointer, return Cb = 1 if pointer >= end
	BCS LA4D8		; if at end go do UNLISTEN
	LDA STATUS		; get I/O status byte
	BNE LA4CC		; branch if error
	LDA (SAL),Y		; get byte from buffer
	JSR DBCIOUT		; output a byte to either the serial or IEEE bus
	JSR STOP		; scan stop key
	BNE LA4D3		; if stop not pressed go increment pointer and loop for next
LA4CC	JSR LA4D8		; go do UNLISTEN
	LDA #ER_STOP		; terminated by STOP key
	SEC			; flag stop
	RTS
LA4D3	JSR WRT62		; increment read/write pointer
	BNE LA4B9		; loop, branch always

LA4D8	JSR DBUNLSN		; command the serial or IEEE bus to UNLISTEN
	JSR LA2DF		; close the serial or IEEE bus device
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; query bus to use
;
; Set carry and overflow status bits according to the current bus configuration
;
; Cb = 0 - serial bus
;    = 1 - IEEE-488 bus
; Vb = 1 - user port (for device 4 only)

QUERYBUS
	PHA			; save .A
	LDA BUSCONF		; get bus config
	CLC
	ADC #$7C		; add constant to transfer b2 into Vb
	LSR			; shift b0 of bus config into Cb
	PLA			; restore .A
	RTS


;***********************************************************************************;
;
; crunch BASIC4 tokens, the crunch BASIC tokens vector is initialized to point here

B4CRNCH
	LDX CHRGOT+1		; get BASIC execute pointer low byte
	LDY CHRGOT+1		; get BASIC execute pointer low byte
LA4EE	JSR COPYBYTE		; copy byte from BUF,.X to BUF,.Y
	BEQ LA558		; branch if was null [EOL]
	CMP #$22		; compare with quote character
	BNE LA4FC
	JSR LA57D		; copy quoted string
	BEQ LA53F
LA4FC	CMP #'A'		; compare with "A"
	BCC LA53F		; branch if <
	CMP #'Z'+1		; compare with character after "Z"
	BCS LA53F		; branch if >=
	STY FBUFPT		; copy save index
	STX XSAV
	LDY #$01
	STY COUNT
	DEX
	LDY #$FF
LA50F	INX			; next table byte
	INY			; next buffer byte
LA511	LDA BUF,X		; get byte from input buffer
	SEC
	SBC B4WORDS,Y		; subtract table byte
	BEQ LA50F		; go compare next if match
	ASL			; shift b7 into Cb, 0 into b0
	BNE LA543		; non-zero so not end marker match, go try the next keyword
	LDY FBUFPT		; restore save index
	LDA COUNT
	CMP #$15		; compare with the number of BASIC4 tokens
	BNE LA53C
	LDX XSAV
	.BYTE $2C		; makes next line BIT $C8E8
LA528	INX			; increment source index
	INY			; increment destination index
	JSR COPYBYTE		; copy byte from BUF,.X to BUF,.Y
	BEQ LA558		; branch if was null [EOL]
	CMP #':'		; compare with ":"
	BEQ LA53F
	CMP #$22		; compare with quote character
	BNE LA528
	JSR LA57D
	BEQ LA528
LA53C	STA BUF,Y
LA53F	INX
	INY
	BNE LA4EE
LA543	LDX XSAV
	INC COUNT		; increment word index (next word)

				; now find end of this word in the table

LA547	INY			; increment table index
	LDA B4WORDS-1,Y		; get table byte
	BPL LA547		; loop if not end of word yet
	LDA B4WORDS,Y		; get byte from keyword table
	BNE LA511		; go test next word if not zero byte, end of table
	LDY FBUFPT		; restore save index
	BPL LA53F
LA556	PLA
	PLA
LA558	JSR CRNCH2		; crunch BASIC2 tokens
	LDX #$00		; clear index
LA55D	LDA BUF,X
	BEQ LA57C		; if EOL then exit
	INX			; increment index
	CMP #$22		; compare with quote character
	BNE LA571
LA567	LDA BUF,X
	BEQ LA57C		; if EOL then exit
	INX			; increment index
	CMP #$22		; compare with quote character
	BNE LA567		; loop until close quote
LA571	CMP #' '
	BCS LA55D
	ADC #$CB		; add token offset
	STA BUF-1,X
	BNE LA55D
LA57C	RTS

LA57D	INX			; increment source index
	INY			; increment destination index
	JSR COPYBYTE		; copy byte from BUF,.X to BUF,.Y
	STA BUF,Y		; ??? copy byte again ???
	BEQ LA556		; branch if was null [EOL]
	CMP #$22		; compare with quote character
	BNE LA57D		; loop if not quote
	RTS


;***********************************************************************************;
;
; copy byte from BUF,.X to BUF,.Y

COPYBYTE
	LDA BUF,X		; get byte
	STA BUF,Y		; set byte
	RTS


;***********************************************************************************;
;
; uncrunch BASIC4 tokens, the uncrunch BASIC tokens vector is initialized to point here

B4QPLOP
	BPL LA5C1		; just go print it if not token byte
	CMP #TK_PI		; compare with the token for PI
	BEQ LA5C1		; just print it if so
	BIT GARBFL		; test the open quote flag
	BMI LA5C1		; just go print character if open quote set
	CMP #$CC		; compare with first BASIC4 token
	BCC LA5C4
	SBC #$CB		; reduce token range to 1 to whatever
	TAX			; copy token # to .X
	STY FORPNT		; save index for line
	LDY #$FF		; start from -1, adjust for pre increment
LA5A8	DEX			; decrement token #
	BEQ LA5B3		; if now found go do printing
LA5AB	INY			; else increment index
	LDA B4WORDS,Y		; get byte from keyword table
	BPL LA5AB		; loop until keyword end marker
	BMI LA5A8		; go test if this is required keyword, branch always
	
				; found keyword, it's the next one
	
LA5B3	INY			; increment keyword table index
	LDA B4WORDS,Y		; get byte from table
	BMI LA5BE		; go restore index, mask byte and print if
				; byte was end marker

	JSR $CB47		; else go print the character
	BNE LA5B3		; go get next character, branch always
LA5BE	JMP $C6EF		; restore index, mask byte and print

LA5C1	JMP $C6F3		; print non-token
LA5C4	JMP $C724		; jump to BASIC2 uncrunch


;***********************************************************************************;
;
; BASIC4 keywords. Each word has b7 set in its last character as an end marker.

B4WORDS
	.BYTE "CONCA",'T'+$80
	.BYTE "DOPE",'N'+$80
	.BYTE "DCLOS",'E'+$80
	.BYTE "RECOR",'D'+$80
	.BYTE "HEADE",'R'+$80
	.BYTE "COLLEC",'T'+$80
	.BYTE "BACKU",'P'+$80
	.BYTE "COP",'Y'+$80
	.BYTE "APPEN",'D'+$80
	.BYTE "DSAV",'E'+$80
	.BYTE "DLOA",'D'+$80
	.BYTE "CATALO",'G'+$80
	.BYTE "RENAM",'E'+$80
	.BYTE "SCRATC",'H'+$80
	.BYTE "DIRECTOR",'Y'+$80
	.BYTE "IEE",'E'+$80
	.BYTE "SERIA",'L'+$80
	.BYTE "PARALLE",'L'+$80
	.BYTE "MONITO",'R'+$80
	.BYTE "MODE",'M'+$80
	.BYTE "DAT",'A'+$80
MONITOR	.BYTE $00		; end marker


;***********************************************************************************;
;
; Action addresses for BASIC4 commands. These are called by pushing the address
; onto the stack and doing an RTS so the actual address - 1 needs to be pushed.

LA645	.BYTE >CONCAT-1		; perform CONCAT (MSB)
	.BYTE >DOPEN-1		; perform DOPEN (MSB)
	.BYTE >DCLOSE-1		; perform DCLOSE (MSB)
	.BYTE >RECORD-1		; perform RECORD (MSB)
	.BYTE >HEADER-1		; perform HEADER (MSB)
	.BYTE >COLLECT-1	; perform COLLECT (MSB)
	.BYTE >BACKUP-1		; perform BACKUP (MSB)
	.BYTE >COPY-1		; perform COPY (MSB)
	.BYTE >APPEND-1		; perform APPEND (MSB)
	.BYTE >DSAVE-1		; perform DSAVE (MSB)
	.BYTE >DLOAD-1		; perform DLOAD (MSB)
	.BYTE >CATALOG-1	; perform CATALOG (MSB)
	.BYTE >RENAME-1		; perform RENAME (MSB)
	.BYTE >SCRATCH-1	; perform SCRATCH (MSB)
	.BYTE >CATALOG-1	; perform DIRECTORY (MSB)
	.BYTE >IEEE-1		; perform IEEE (MSB)
	.BYTE >SERIAL-1		; perform SERIAL (MSB)
	.BYTE >PARALLEL-1	; perform PARALLEL (MSB)
	.BYTE >MONITOR-1	; perform MONITOR (MSB)
	.BYTE >MODEM-1		; perform MODEM (MSB)
	.BYTE >DATA-1		; perform DATA (MSB)

LA65A	.BYTE <CONCAT-1		; perform CONCAT (LSB)
	.BYTE <DOPEN-1		; perform DOPEN (LSB)
	.BYTE <DCLOSE-1		; perform DCLOSE (LSB)
	.BYTE <RECORD-1		; perform RECORD (LSB)
	.BYTE <HEADER-1		; perform HEADER (LSB)
	.BYTE <COLLECT-1	; perform COLLECT (LSB)
	.BYTE <BACKUP-1		; perform BACKUP (LSB)
	.BYTE <COPY-1		; perform COPY (LSB)
	.BYTE <APPEND-1		; perform APPEND (LSB)
	.BYTE <DSAVE-1		; perform DSAVE (LSB)
	.BYTE <DLOAD-1		; perform DLOAD (LSB)
	.BYTE <CATALOG-1	; perform CATALOG (LSB)
	.BYTE <RENAME-1		; perform RENAME (LSB)
	.BYTE <SCRATCH-1	; perform SCRATCH (LSB)
	.BYTE <CATALOG-1	; perform DIRECTORY (LSB)
	.BYTE <IEEE-1		; perform IEEE (LSB)
	.BYTE <SERIAL-1		; perform SERIAL (LSB)
	.BYTE <PARALLEL-1	; perform PARALLEL (LSB)
	.BYTE <MONITOR-1	; perform MONITOR (LSB)
	.BYTE <MODEM-1		; perform MODEM (LSB)
	.BYTE <DATA-1		; perform DATA (LSB)


;***********************************************************************************;
;
; Disk command templates. Describes the command letter, punctuation and parameter
; order for each DOS command.

CMDTEMPL
	.BYTE '$',$D1,':'
	.BYTE $F1,$D1,':'
	.BYTE $F1,',',$E1
	.BYTE ',',$E0,$D1
	.BYTE ':',$F1,','
	.BYTE $41,'N',$D1
	.BYTE ':',$F1,','
	.BYTE $D0,'V',$D1
	.BYTE 'D',$D2,'='
	.BYTE $D1,'C',$D2
	.BYTE ':',$F2,'='
	.BYTE $D1,':',$F1
	.BYTE 'C',$D2,':'
	.BYTE $F2,'=',$D2
	.BYTE ':',$F2,','
	.BYTE $D1,':',$F1
	.BYTE 'R',$D1,':'
	.BYTE $F2,'=',$D1
	.BYTE ':',$F1,'S'
	.BYTE $D1,':',$F1


;***********************************************************************************;
;
; start new BASIC code, the start new BASIC code vector is initialized to point here

B4GONE
	LDA CHRGOT+1		; get BASIC execute pointer low byte
	PHA			; save it
	LDA CHRGOT+2		; get BASIC execute pointer high byte
	PHA			; save it
LA6B1	JSR CHRGET		; increment and scan memory
	BEQ LA6D2		; branch if no more chrs
	CMP #$22		; compare with quote character
	BNE LA6C3		; branch if not quote
LA6BA	JSR CHRGET		; increment and scan memory
	BEQ LA6D2		; branch if no more chrs
	CMP #$22		; compare with quote character
	BNE LA6BA		; loop until end quote
LA6C3	CMP #'D'		; compare character with "D"
	BNE LA6B1
	LDY #$01
	LDA (CHRGOT+1),Y	; get next BASIC byte
	CMP #'S'		; compare next character with "S"
	BNE LA6B1		; branch if not "DS" variable
	LDA STVALID		; get disk status valid
LA6D2	BEQ LA711		; if disk status valid continue parsing
	LDA #$00		; clear .A
	STA STATUS		; clear I/O status byte
	LDA FA			; get device number
	CMP #$05
	BCS LA6E2
	LDA #$08		; set device to 8, first disk
	STA FA			; set device number
LA6E2	JSR DBTALK		; command a device on either the serial or IEEE bus to TALK
	LDA STATUS		; get I/O status byte
	BEQ LA6F9		; if no error read disk status
LA6E9	LDX #'1'
	STX CMDBUF		; store '1'
	DEX
	STX CMDBUF+1		; store '0'
	LDX #$00
	STX CMDBUF+2		; store NUL
	BEQ LA713		; set value of DS$, branch always


;***********************************************************************************;
;
; read disk status from command channel

LA6F9	LDA #$6F		; set secondary address to 15 (command channel)
	JSR DBTKSA		; send secondary address after TALK
	LDX #$00		; clear index
LA700	JSR DBACPTR		; input a byte from either the serial or IEEE bus
	STA CMDBUF,X		; save byte in buffer
	INX			; increment index
	LDA STATUS		; get I/O status byte
	BEQ LA700		; loop while no error
	JSR DBUNTLK		; command the serial or IEEE bus to UNTALK
	DEX			; decrement buffer length
	BEQ LA6E9

LA711	BEQ LA762		; branch always


;***********************************************************************************;
;
; store disk status in DS$

LA713	TXA			; copy string length
	PHA			; save string length
	LDA #'D'		; set first character to "D"
	STA VARNAM		; set variable name 1st character
	LDA #'S'+$80		; set second character to "S$"
	STA VARNAM+1		; set variable name 2nd character
	JSR FNDVAR		; find or create variable
	PLA			; restore string length
	PHA			; save string length
	JSR ALCSPAC		; make space in string memory for string
	LDY #$02
	LDA FRESPC+1		; get string utility ptr high byte
	STA (VARPNT),Y		; set string location high byte
	LDA #>CMDBUF		; get status message high byte
	STA CHRGOT+2		; save BASIC execute pointer high byte
	DEY			; decrement index
	LDA FRESPC		; get save string utility ptr low byte
	STA (VARPNT),Y		; set string location low byte
	LDA #<CMDBUF		; get status message low byte
	STA CHRGOT+1		; save BASIC execute pointer low byte
	DEY			; decrement index
	STY STVALID		; set disk status valid
	PLA			; restore string length
	STA (VARPNT),Y		; save string length
	TAY			; copy string length to .Y
LA740	DEY			; decrement index
	BMI LA74A		; if empty string store error code
	LDA CMDBUF,Y		; get character
	STA (FRESPC),Y		; save character
	BNE LA740		; loop while characters left


;***********************************************************************************;
;
; store disk error code in DS

LA74A	LDA #'S'		; set second character to "S"
	STA VARNAM+1		; set variable name 2nd character
	JSR FNDVAR		; find or create variable
	JSR FRMEVL		; evaluate expression
	ASL FAC1+FAC_MANT	; shift FAC1 mantissa
	LSR FAC1+FAC_MANT	; clear FAC1 sign bit
	LDY #$04		; set index
LA75A	LDA !FAC1,Y		; get FAC1,.Y
	STA (VARPNT),Y		; save byte to variable
	DEY			; decrement index
	BPL LA75A		; loop until value copied
LA762	PLA			; restore previous BASIC execute pointer high byte
	STA CHRGOT+2		; save BASIC execute pointer high byte
	PLA			; restore previous BASIC execute pointer low byte
	STA CHRGOT+1		; save BASIC execute pointer low byte
	JSR CHRGET		; increment and scan memory
	.BYTE $2C		; makes next line BIT $6868
LA76C	PLA
	PLA
	JSR CHRGOT		; scan memory
	BEQ LA788		; branch if no more chrs
	CMP #TK_IF		; compare with token for IF
	BNE LA779
	LDA #$E0
LA779	SEC
	SBC #$CC		; subtract first BASIC4 token
	BCC LA788		; branch if less than the first token
	CMP #$15		; compare with number of BASIC4 tokens
	BCS LA788		; branch if greater
	JSR LA78E		; go interpret token
	JMP NEWSTT		; go do interpreter inner loop

LA788	JSR CHRGOT		; scan memory
	JMP $C7E7		; go interpret BASIC code and loop

LA78E	TAY			; copy to index
	LDA LA645,Y		; get vector high byte
	PHA			; push on stack
	LDA LA65A,Y		; get vector low byte
	PHA			; push on stack
	JMP CHRGET		; increment and scan memory and return. the return in
				; this case calls the command code, the return from
				; that will eventually return to the interpreter inner
				; loop


;***********************************************************************************;
;
; perform DATA

DATA
	JSR FRMEVL		; evaluate expression
	JSR CHRGOT		; scan memory
	CMP #TK_GOTO		; compare with token for GOTO
	BEQ LA7A9		; branch if GOTO
	LDA #TK_THEN		; set THEN token
	JSR SYNCHR		; scan for CHR$(.A), else do syntax error
LA7A9	LDA FAC1+FAC_EXPT	; get FAC1 exponent
	BEQ LA7B5		; branch if FAC1 = $00
	JSR CHRGOT		; scan memory
	BCS LA76C		; branch if not numeric character, is variable or keyword
	JMP GOTO		; perform GOTO
LA7B5	JMP REM			; REM the rest of the line


;***********************************************************************************;
;
; perform IEEE

IEEE
	BNE LA80F		; exit if following byte to allow syntax error
	LDA #$03		; set IEEE bus
	BNE LA7CB		; branch to set bus config


;***********************************************************************************;
;
; perform SERIAL

SERIAL
	BNE LA80F		; exit if following byte to allow syntax error
	LDA #$02		; set serial bus
	BNE LA7CB		; branch to set bus config


;***********************************************************************************;
;
; perform PARALLEL

PARALLEL
	BNE LA80F		; exit if following byte to allow syntax error
	LDA BUSCONF		; get bus config
	ORA #$04		; OR printer to user port bit
LA7CB	STA BUSCONF		; set bus config
	STA STVALID		; reset disk status valid
	RTS


;***********************************************************************************;
;
; parse file name and, optionally, device and drive number

PARSEFNAME
	JSR PARSEPARAM		; parse disk command parameters
	AND #$E6		; mask illegal parameters
	BNE LA80F		; if any set do syntax error
LA7D9	LDA PARSEST		; get parse state
	LSR			; shift primary file name flag into Cb
	BCC LA80F		; if file name not present do syntax error
	ROL			; restore parse state
	RTS

LA7E0	AND #$C4		; mask illegal parameters
	BNE LA80F		; if any set do syntax error
	LDA PARSEST		; get parse state
LA7E6	AND #$03		; mask primary and secondary file name
	CMP #$03		; test both are present
LA7EA	BNE LA80F		; if either missing do syntax error
	LDA PARSEST		; get parse state
	RTS


;***********************************************************************************;
;
; parse file name and drive number

PARSENAMD
	JSR PARSEPARAM		; parse disk command parameters
	AND #$05		; mask primary drive number and file name
	CMP #$05		; test both are present
	JMP LA7EA		; do syntax error if either are missing


;***********************************************************************************;
;
; set secondary address to free value

SETFREESA
	LDY #$61		; set secondary address to $01
LA7FB	INY			; increment secondary address
	TYA			; copy secondary address
	LDX LDTND		; get open file count
LA7FF	DEX			; decrement count to give index
	BMI LA809		; exit if no files
	CMP SAT,X		; compare with secondary address table
	BEQ LA7FB		; loop if secondary address found
	BNE LA7FF		; loop through whole table
LA809	STY SA			; set secondary address
	RTS

LA80C	JMP LABE5		; do illegal quantity error
LA80F	JMP $CF08		; do syntax error then warm start


;***********************************************************************************;
;
; perform RECORD

RECORD
	LDA #$01		; set byte position to first byte
	STA CMDBUF+4		; set record byte position
	JSR CHRGOT		; scan memory
	LDA #'#'		; get # token
	JSR SYNCHR		; scan for CHR$(.A), else do syntax error
	JSR LAC36		; evaluate byte expression, result in .X
	TXA			; copy logical file
	BEQ LA80C		; if = 0 do illegal quantity error
	STX LA			; set logical file
	JSR COMCHK		; scan for ",", else do syntax error
	BEQ LA80F		; if no following value do syntax error
	BCC LA83D		; branch if numeric
	JSR LPACHK		; scan for "(", else do syntax error
	JSR FRMEVL		; evaluate expression
	JSR MAKADR		; convert FAC1 to integer in temporary integer
	JSR RPACHK		; scan for ")", else do syntax error
	JMP LA843		; branch always
LA83D	JSR FRMEVL		; evaluate expression
	JSR MAKADR		; convert FAC1 to integer in temporary integer
LA843	JSR CHRGOT		; scan memory
	BEQ LA85E		; branch if no more chrs
	JSR COMCHK		; scan for ",", else do syntax error
	BEQ LA80F		; if no following value do syntax error
	JSR LAC36		; evaluate byte expression, result in .X
	TXA			; copy record byte position
	BEQ LA80C		; if = 0 do illegal quantity error
	INX			; increment to test ..
	BEQ LA80C		; if = $FF do illegal quantity error
	STA CMDBUF+4		; set record byte position
	JSR CHRGOT		; scan memory
	BNE LA80F		; branch if not null
LA85E	LDA LA			; get logical file
	JSR $F3D4		; find file .A
	BEQ LA86A		; branch if the file is found
	LDX #ER_FNOTOPEN	; file not found
	JMP ERROR		; do error #.X then warm start
LA86A	JSR SETFLCH		; set file details from table,X
	LDA #'P'		; "P" (record position)
	STA CMDBUF		; set record position command
	LDA SA			; get secondary address
	STA CMDBUF+1		; set channel number
	LDA LINNUM		; get temporary integer low byte
	STA CMDBUF+2		; set record number low byte
	LDA LINNUM+1		; get temporary integer high byte
	STA CMDBUF+3		; set record number high byte
	LDX #$05		; set command length
	JSR LAA69		; setup buffer address and length
	JMP LA954		; send on command channel


;***********************************************************************************;
;
; Prompt the user to proceed with a destructive action. If negative then the calling
; subroutine is terminated.

RUSURE
	LDA CHRGOT+2		; save BASIC execute pointer high byte
	CMP #$02		; compare with $02xx
	BNE LA8AC		; if not immediate mode return
	LDY #$0D		; set index to length of prompt
LA891	LDA LA8AD,Y		; get character from prompt
	JSR SCRNOUT		; output character
	DEY			; decrement index
	BPL LA891		; loop until all done
	JSR FCHRIN		; get input from keyboard
	PHA			; save key
LA89E	JSR FCHRIN		; get input from keyboard
	CMP #$0D		; compare with [CR]
	BNE LA89E		; loop until [CR]
	PLA			; restore first key
	CMP #'Y'		; compare key with "Y"
	BEQ LA8AC		; branch if accepted
	PLA			; dump return address low byte
	PLA			; dump return address high byte
LA8AC	RTS

LA8AD	.BYTE "?ERUS UOY ERA",$0D


;***********************************************************************************;
;
; close all logical files on the current device

CLOSEALL
	LDA FA			; get device number
	LDX LDTND		; get open file count
LA8BF	DEX			; decrement count to give index
	BMI LA8AC		; exit if no files
	CMP FAT,X		; compare device number with table device number
	BNE LA8BF		; loop if no match
	LDA LAT,X		; get logical file number from table logical file number
	JSR CLOSE		; close a specified logical file
	JMP CLOSEALL		; loop for all logical files


;***********************************************************************************;
;
; perform DCLOSE

DCLOSE
	JSR PARSEPARAM		; parse disk command parameters
	AND #$F3		; mask illegal parameters
	BNE LA941		; if any set do syntax error
	LDA LA			; get logical file
	BEQ CLOSEALL		; if no logical file close all logical files
	JMP CLOSE		; close specific logical file


;***********************************************************************************;
;
; perform DOPEN

DOPEN
	JSR PARSENAMD		; parse file name and drive number
	AND #$22		; mask illegal parameters
	BNE LA906		; if any set do syntax error
	JSR SETFREESA		; set secondary address to free value
	LDX #$00
	BIT PARSEST		; test parse state
	BPL LA8F4		; branch if replace flag not set
	LDA #'@'		; set replace file
	STA CMDBUF		; store in file name
	INX			; increment file name length
LA8F4	LDY #$03		; set template index
	TYA			; set template length
	BVC LA8FB
	LDA #$07
LA8FB	JSR LAA06
	JMP $E1BE		; open a logical file, handle I/O error


;***********************************************************************************;
;
; perform APPEND

APPEND
	JSR PARSENAMD		; parse file name and drive number
	AND #$E2		; mask illegal parameters
LA906	BNE LA923		; if any set do syntax error
	JSR SETFREESA		; set secondary address to free value
	LDY #$0A		; set template index
	LDA #$05		; set template length
	INX
	BEQ LA8FB		; branch always


;***********************************************************************************;
;
; perform SCRATCH

SCRATCH
	JSR PARSEFNAME		; parse file name etc
	JSR RUSURE		; prompt for confirmation
	LDY #$37
	BNE LA94F		; perform disk command, branch always


;***********************************************************************************;
;
; perform HEADER

HEADER
	JSR PARSEFNAME		; parse disk name etc
	AND #$11		; mask mandatory parameters
	CMP #$11		; test all are present
LA923	BNE LA941		; if any missing do syntax error
	JSR CLOSEALL		; close all logical files
	JSR RUSURE		; prompt for confirmation
	LDY #$0F		; set template index
	LDA #$04		; set template length
	LDX DISKID		; get first disk identifier character
	BEQ LA951		; if not set perform disk command
	LDA #$06
	BNE LA951		; perform disk command, branch always


;***********************************************************************************;
;
; perform BACKUP

BACKUP
	JSR PARSEPARAM		; parse disk command parameters
	AND #$30		; mask mandatory parameters
	CMP #$30		; test all are present
	BEQ LA944		; if all present continue
LA941	JMP $CF08		; do syntax error then warm start
LA944	LDA PARSEST		; get parse state
	AND #$C7		; mask illegal parameters
	BNE LA941		; if any set do syntax error
	JSR CLOSEALL		; close all logical files
	LDY #$17		; set template index
LA94F	LDA #$04		; set template length
LA951	JSR LAA04		; build disk command
LA954	LDA #$6F		; set secondary address to $0F
	STA SA			; set secondary address
	JSR DBSERNAME		; send secondary address and filename
	RTS


;***********************************************************************************;
;
; perform COPY

COPY
	JSR PARSEPARAM		; parse disk command parameters
	AND #$30
	CMP #$30
	BNE LA96B
	LDA PARSEST		; get parse state
	AND #$C7
	BEQ LA970
LA96B	LDA PARSEST		; get parse state
	JSR LA7E0
LA970	LDY #$1B		; set template index
LA972	LDA #$08		; set template length
	BNE LA951		; branch always


;***********************************************************************************;
;
; perform CONCAT

CONCAT
	JSR PARSEPARAM		; parse disk command parameters
	JSR LA7E0
	LDY #$23		; set template index
	LDA #$0C
	BNE LA951		; branch always


;***********************************************************************************;
;
; perform COLLECT

COLLECT
	JSR PARSEPARAM		; parse disk command parameters
	AND #$E7		; mask illegal parameters
	BNE LA941		; if any set do syntax error
	JSR CLOSEALL		; close all logical files
	LDY #$15		; set template index
	LDX #$01
	LDA PARSEST		; get parse state
	AND #$10
	BEQ LA997
	INX
LA997	TXA			; set template length
	BNE LA951		; branch always


;***********************************************************************************;
;
; perform RENAME

RENAME
	JSR PARSEPARAM		; parse disk command parameters
	JSR LA7E6
	AND #$E4		; mask illegal parameters
LA9A2	BNE LA941		; if any set do syntax error
	LDY #$2F		; set template index
	BNE LA972		; branch always


;***********************************************************************************;
;
; perform DLOAD

DLOAD
	JSR PARSEPARAM		; parse disk command parameters
	JSR LA7D9
	AND #$E6		; mask illegal parameters
	BNE LA9A2		; if any set do syntax error
	LDY #$03		; set template index
	TYA			; set template length
	JSR LAA04		; build disk command
	LDX #$00		; flag load
	STX VERCHK		; set load/verify flag
	STX SA			; set secondary address
	JMP $E16C		; jump to BASIC load


;***********************************************************************************;
;
; perform DSAVE

DSAVE
	JSR PARSEPARAM		; parse disk command parameters
	JSR LA7D9
	AND #$66		; mask illegal parameters
	BNE LA9A2		; if any set do syntax error
	TAX			; clear .X
	LDY #$03		; set template index
	BIT PARSEST		; test parse state
	BPL LA9D8		; branch if replace flag not set
	LDA #'@'		; set replace file
	STA CMDBUF		; store in file name
	INX			; increment file name length
LA9D8	TYA			; set template length
	JSR LAA06
	JMP $E156		; jump to BASIC save


;***********************************************************************************;
;
; copy primary file name to command buffer

COPYFNAME1
	LDA FNLEN
	STA CMP0
	LDA FNADR
	STA EAL
	LDA FNADR+1
	STA EAL+1


;***********************************************************************************;
;
; copy secondary file name to command buffer

COPYFNAME2
	TYA
	PHA
	LDY CMP0
	BEQ LA9FF
	LDY #$00		; clear index
LA9F3	LDA (EAL),Y
	STA CMDBUF,X
	INX			; increment destination index
	INY			; increment source index
	CPY CMP0		; compare source index with length
	BNE LA9F3		; loop while not equal
	.BYTE $24		; makes next line BIT $CA
LA9FF	DEX
	PLA
	TAY
	SEC
	RTS


;***********************************************************************************;
;
; build disk command

LAA04	LDX #$00		; set command length
LAA06	STA TEMPLEN		; set template length
LAA09	DEC TEMPLEN		; decrement template length
	BMI LAA69		; if non-zero setup buffer address and length
	INY			; increment index
	LDA CMDTEMPL,Y		; get template character
	BPL LAA63		; if normal character store in buffer
	CMP #$F1		; compare with primary file name token
	BNE LAA1B		; branch if no match
	JSR COPYFNAME1		; copy primary file name to buffer
LAA1B	CMP #$F2		; compare with secondary file name token
	BNE LAA22		; branch if no match
	JSR COPYFNAME2		; copy primary file name to buffer
LAA22	CMP #$E0		; compare with record length token
	BNE LAA2B		; branch if no match
	LDA RECLEN		; get record length
	BNE LAA63		; store in command, branch always
LAA2B	CMP #$D0		; compare with disk name token
	BNE LAA3E		; branch if no match
	LDA DISKID		; get first disk identifier character
	STA CMDBUF,X		; store in command
	INX			; increment index
	LDA DISKID+1		; get second disk identifier character
	STA CMDBUF,X		; store in command
	INX			; increment index
	TXA			; copy index to .A
LAA3E	CMP #$E1
	BNE LAA53		; branch if no match
	LDA RECLEN		; get record length
	BNE LAA4F		; if set store length indicator
	LDA #'S'		; SEQ file type
	STA RECLEN
	LDA #'W'		; write file mode
	.BYTE $2C		; makes next line BIT $4CA9
LAA4F	LDA #'L'		; "L", record length
	BCS LAA63		; store in command, branch always
LAA53	CMP #$D1		; compare with primary drive number
	BNE LAA5B
	LDA DRIVENO1		; get primary drive number
	BPL LAA61		; branch always
LAA5B	CMP #$D2		; compare with secondary drive number
	BNE LAA09		; if no match process next template token
	LDA DRIVENO2		; get secondary drive number
LAA61	ORA #'0'		; convert to numeric character
LAA63	STA CMDBUF,X		; store in command
	INX			; increment index
	BNE LAA09		; branch always

LAA69	STX FNLEN		; set buffer length
	LDA #<CMDBUF		; get command buffer low byte
	STA FNADR		; set buffer pointer low byte
	LDA #>CMDBUF		; get command buffer high byte
	STA FNADR+1		; set buffer pointer high byte
	RTS


;***********************************************************************************;
;
; parse disk command parameters

PARSEPARAM
	LDX #$00		; clear .X
	STX PARSEST		; clear parse state
	STX LA			; clear logical file
	STX RECLEN		; clear record length/disk identifier flag
	STX DRIVENO2		; clear secondary drive number
	STX DRIVENO1		; clear primary drive number
	STX FNLEN		; clear primary file name length
	STX CMP0		; clear secondary file name length
	STX DISKID		; clear first disk identifier character
	STX STATUS		; clear I/O status byte
	LDX #$08		; set device to 8, first disk
	STX FA			; set device number
	JSR CHRGOT		; scan memory
	BEQ LAAFA		; if no more chrs get parse state and exit

; parse primary parameters

LAA93	CMP #'#'		; compare with "#" character
	BEQ EVALLFN		; evaluate logical file parameter
	CMP #'W'
	BEQ EVALRECW		; evaluate record length/write parameter
	CMP #'L'		; compare with "L", record length
	BEQ EVALRECW		; evaluate record length/write parameter
	CMP #'R'
	BNE LAAA9
LAAA3	JSR CHRGET		; increment and scan memory
LAAA6	JMP LAB6B
LAAA9	CMP #'D'		; compare with "D", drive number
	BEQ LAB08		; evaluate primary drive number parameter
	CMP #TK_ON		; compare with ON token
	BEQ LAAFE		; evaluate device number parameter
	CMP #'U'		; compare with "U", device number
	BEQ LAB03		; evaluate device number parameter
	CMP #'I'		; compare with "I", disk identifier
	BEQ LAB1D		; evaluate disk identifier
	CMP #$22		; compare with quote character
	BEQ LAAFC		; evaluate primary file name parameter
	CMP #'('		; compare with "("
	BEQ LAAFC		; evaluate primary file name parameter
LAAC1	JMP $CF08		; do syntax error then warm start


;***********************************************************************************;
;
; evaluate logical file parameter

EVALLFN
	LDA PARSEST		; get parse state
	AND #$04		; mask logical file bit
	BNE LAAC1		; if already set do syntax error
	JSR GETNUMERIC		; get byte expression in .X
	TXA			; copy value to .A
	BEQ LAAD6		; if zero do illegal quantity error
	STX LA			; set logical file
	LDA #$04		; set logical file bit
	BNE LAAF5		; update parse state, branch always

LAAD6	JMP LABE5		; do illegal quantity error


;***********************************************************************************;
;
; evaluate record length/write parameter

EVALRECW
	BIT PARSEST		; test parse state
	BVS LAAC1		; if record length/write parsed then do syntax error
	CMP #'W'		; compare with "W"
	BNE LAAE7		; if not "W" then evaluate record length parameter
	JSR CHRGET		; increment and scan memory
	JMP LAAF3		; set record length/write bit in parse state and continue


;***********************************************************************************;
;
; evaluate record length parameter

LAAE7	JSR GETNUMERIC		; get byte expression in .X
	TXA			; copy value to .A
	BEQ LAAD6		; if zero do illegal quantity error
	INX			; increment to roll over to zero
	BEQ LAAD6		; if was 255 do illegal quantity error
	STA RECLEN		; save record length
LAAF3	LDA #$40		; set record length/write bit
LAAF5	JSR LAC00		; update parse state
	BNE LAAA6		; branch always

LAAFA	BEQ LAB70		; get parse state and exit
LAAFC	BEQ LAB3C		; evaluate primary file name parameter

LAAFE	JSR GETDEVNUM		; evaluate device number parameter
	BNE LAB6B		; branch always


;***********************************************************************************;
;
; evaluate device number parameter

LAB03	JSR LABF1
	BNE LAB6B		; branch always


;***********************************************************************************;
;
; evaluate primary drive number parameter

LAB08	LDA PARSEST		; get parse state
	AND #$10		; mask primary drive number bit
	BNE LAAC1		; if set do syntax error
	JSR GETNUMERIC		; get byte expression in .X
	CPX #$02		; compare with max drive + 1
	BCS LAAD6		; if >= do illegal quantity error
	STX DRIVENO1		; set primary drive number
	STX DRIVENO2		; set secondary drive number
	LDA #$10		; set primary drive number parameter bit
	BNE LAAF5		; update parse state, branch always


;***********************************************************************************;
;
; evaluate disk identifier parameter

LAB1D	LDA RECLEN		; get record length/disk identifier flag
	BNE LAAC1		; if either already parsed do syntax error
	JSR CHRGET		; increment and scan memory
	BEQ LAB9C		; branch if no more chrs
	STA DISKID		; set first disk identifier character
	JSR CHRGET		; increment and scan memory
	BEQ LAB9C		; branch if no more chrs
	STA DISKID+1		; set second disk identifier character
	LDA #$FF		; set invalid record length
	STA RECLEN		; set record length/disk identifier flag
	JMP LAAA3		; continue parsing

LAB3A	BEQ LAAFE		; evaluate device number parameter, branch always


;***********************************************************************************;
;
; evaluate primary file name parameter

LAB3C	LDA PARSEST		; get parse state
	LSR			; shift primary file name bit into Cb
	JSR GETSTRING		; evaluate string
	STA FNLEN		; set file name length
	STA TEMPLEN		; set string length
	LDA #<TEMPSTR
	STA FNADR		; set file name pointer low byte
	LDA #>TEMPSTR
	STA FNADR+1		; set file name pointer high byte
	LDY #$00		; clear index
	LDA (INDEX),Y		; get string character
	CMP #'@'		; compare with replace command
	BNE LAB5B		; branch if not replace
	INC FNADR		; increment file name pointer low byte
	DEC FNLEN		; decrement file name length
LAB5B	STA TEMPSTR,Y		; store string character
	INY			; increment index
	LDA (INDEX),Y		; get string character
	CPY TEMPLEN		; compare with string length
	BCC LAB5B		; loop until equal
	LDA #$01		; set primary file name bit
	JSR LAC00		; update parse state
LAB6B	JSR CHRGOT		; scan memory
	BNE LAB73		; branch if not null
LAB70	LDA PARSEST		; get parse state
	RTS

LAB73	CMP #','
	BNE LAB7D
	JSR CHRGET		; increment and scan memory
	JMP LAA93		; continue parsing primary parameters


;***********************************************************************************;
;
; parse secondary parameters

LAB7D	CMP #TK_ON		; compare with ON token
	BEQ LAB3A		; evaluate device number parameter
	CMP #TK_TO		; compare with TO token
	BNE LAB9C		; if not TO do syntax error
LAB85	JSR CHRGET		; increment and scan memory
	CMP #'D'		; compare with "D", drive number
	BEQ LAB9F		; evaluate secondary drive number parameter
	CMP #TK_ON		; compare with ON token
	BEQ LABB5		; if ON get device number
	CMP #'W'		; compare with "W"
	BEQ LABBA
	CMP #$22		; compare with quote character
	BEQ LABBF		; evaluate secondary file name parameter
	CMP #'('		; compare with "("
	BEQ LABBF		; evaluate secondary file name parameter
LAB9C	JMP $CF08		; do syntax error then warm start


;***********************************************************************************;
;
; evaluate secondary drive number parameter

LAB9F	LDA PARSEST		; get parse state
	AND #$20		; mask secondary drive number bit
	BNE LAB9C		; if set do syntax error
	JSR GETNUMERIC		; get byte expression in .X
	CPX #$02		; compare with max drive + 1
	BCS LABE5		; if >= do illegal quantity error
	STX DRIVENO2		; set secondary drive number
	LDA #$20		; set secondary drive number bit
	JSR LAC00		; update parse state
	BNE LABD1		; continue parsing, branch always

LABB5	JSR GETDEVNUM		; evaluate device number parameter
	BNE LABD1		; continue parsing, branch always

LABBA	JSR LABF1
	BNE LABD1		; continue parsing, branch always


;***********************************************************************************;
;
; evaluate secondary file name parameter

LABBF	LDA PARSEST		; get parse state
	LSR			; shift secondary file name bit ..
	LSR			; .. into Cb
	JSR GETSTRING		; evaluate string
	STA CMP0		; set string length
	STX EAL			; set string pointer low byte
	STY EAL+1		; set string pointer high byte
	LDA #$02		; set secondary file name bit
	JSR LAC00		; update parse state
LABD1	JSR CHRGOT		; scan memory
	BEQ LAB70		; if no more chrs get parse state and exit
	CMP #','		; compare with ","
	BEQ LAB85
	CMP #TK_ON		; compare with ON token
	BEQ LABB5		; if ON get device number
	CMP #'W'		; compare with "W"
	BEQ LABBA
LABE2	JMP $CF08		; do syntax error then warm start

LABE5	LDX #ER_ILLQUAN		; illegal quantity error
LABE7	JMP ERROR		; do error #.X then warm start


;***********************************************************************************;
;
; evaluate device number parameter

GETDEVNUM
	JSR CHRGET		; increment and scan memory
	CMP #'U'		; compare with "U", device number parameter
LABEF	BNE LABE2		; if not "U" do syntax error
LABF1	JSR GETNUMERIC		; get byte expression in .X
	CPX #$20		; compare with max device + 1
	BCS LABE5		; if device number too large do illegal quantity error
	CPX #$04		; compare with first device number
	BCC LABE5		; if too small do illegal quantity error
	STX FA			; set device number
	LDA #$08		; set device number bit
LAC00	ORA PARSEST		; OR in new bit
	STA PARSEST		; set parse state
	RTS


;***********************************************************************************;
;
; evaluate string, returns with .A = length, .X = pointer low byte, .Y = pointer high byte

GETSTRING
	BCS LABE2		; if already parsed do syntax error
	JSR FRMEVL		; evaluate expression
	JSR DELST		; evaluate string
	TAX			; copy length to .X
	BEQ LABE5		; if empty do illegal quantity error
	CMP #$12
	BCS LAC20		; if >= do string too long
	LDY #$00		; clear index
	LDA (INDEX),Y		; get first character of string
	CMP #'@'		; compare with replace command
	BEQ LAC24		; branch if replace
	CPX #$11
	BCC LAC2D		; branch if length ok
LAC20	LDX #ER_STR2LONG	; string too long
	BNE LABE7		; do error #.X then warm start
LAC24	LDA PARSEST		; get parse state
	BMI LABE2		; do error if replace already processed
	LDA #$80		; set replace bit
	JSR LAC00		; update parse state
LAC2D	TXA			; copy length to .A
	LDX INDEX		; get string pointer low byte
	LDY INDEX+1		; get string pointer high byte
	RTS


;***********************************************************************************;
;
; evaluate byte expression, result in .X

GETNUMERIC
	JSR CHRGET		; increment and scan memory
LAC36	BEQ LABE2		; if no more chrs do syntax error
	BCC LAC43		; branch if numeric
	JSR LPACHK		; scan for "(", else do syntax error
	JSR $D79E		; get byte parameter
	JMP RPACHK		; scan for ")", else do syntax error
LAC43	JMP $D79E		; get byte parameter

LAC46	JMP LAD1F		; close logical file and restore previous output device

	
;***********************************************************************************;
;
; perform CATALOG/DIRECTORY

CATALOG
	LDA LA			; get logical file
	STA EAL+1		; save current logical file
	JSR PARSEPARAM		; parse disk command parameters
	AND #$E6		; mask illegal parameters
	BNE LABEF		; if any set do syntax error
	STA CMP0		; clear header/directory entry flag
	LDY #$FF		; set template index
	LDX #$01		; set template length
	LDA PARSEST		; get parse state
	LSR			; shift primary file name bit into Cb
	BCS LAC65		; branch if file name given
	AND #$08		; test primary drive number bit ($10 shifted right)
	BEQ LAC67		; branch if no drive number given
	INX			; increment template length
	.BYTE $2C		; makes next line BIT $04A2
LAC65	LDX #$04
LAC67	TXA			; set template lenth
	JSR LAA04		; build disk command
	LDA DFLTO		; get output device number
	STA EAL			; save current output device
	LDA #$60		; set secondary address to $00
	STA SA			; set secondary address
	STA LA			; set logical file
	JSR DBUNLSN		; command the serial or IEEE bus to UNLISTEN
	LDA STATUS		; get I/O status byte
	BEQ LAC81		; branch if no error
	LDX #ER_DEVNOTP		; device not present
	JMP ERROR		; do error #.X then warm start

LAC81	JSR B4OPEN		; open a logical file
	LDY #$03		; consume 3 words, load address, line link, line number

LAC86	LDX #$60		; set logical file to $60
	JSR B4CHKIN		; open a channel for input
LAC8B	JSR DBACPTR		; get number of blocks low byte
	STA FAC1+FAC_MANT+1	; save byte
	LDX STATUS		; get I/O status byte
	BNE LAC46		; if any error close logical file and restore previous output device
	JSR DBACPTR		; get number of blocks high byte
	STA FAC1+FAC_MANT	; save byte
	DEY			; decrement word count
	BNE LAC8B		; loop until done
	JSR CLRCHN		; close input and output channels
	JSR RESTOUT		; restore previous output device
	LDX FAC1+FAC_MANT+1
	JSR $DDD1		; print number of blocks as unsigned integer
	JSR PRINTSP		; print space
	JSR B4CLRCHN		; close input and output channels
	LDA #$12		; file name + two quotes
	STA FNLEN		; set file name length
	LDX EAL			; get saved output device
	CPX #$03		; compare with screen
	BNE LACC0		; if not screen skip column check
	LDA PNTR		; get cursor column
	CMP #$04		; compare with column 4
	BCS LACC0		; if beyond column 3 already skip column reset
	DEX			; decrement column to 2
	STX PNTR		; set cursor column

LACC0	LDX #$60		; set logical file to $60
	JSR B4CHKIN		; open a channel for input
	JSR DBACPTR		; input a byte from either the serial or IEEE bus
	PHA			; save byte
	JSR B4CLRCHN		; close input and output channels
	LDA STATUS		; get I/O status byte
	BNE LAD1E		; if any error close and exit
	PLA			; restore byte
	BEQ LAD14		; if EOL print [CR] and continue
	LDX EAL			; get saved output device
	CPX #$03		; compare with screen
	BNE LAD00		; if not screen skip entry transformation
	CMP #' '		; compare with space
	BNE LACE3		; if not space it is the file name
	LDX PNTR		; get cursor column
	CPX #$02
	BEQ LACC0		; if padding at start of line then skip
LACE3	LDX FNLEN		; get file name length
	BMI LACC0		; beyond file name, skip rest of entry
	BNE LACFE		; part of file name, print it
	LDX CMP0		; get header/directory entry flag
	BEQ LAD00		; if in header just print it
	CMP #' '		; compare with space
	BEQ LACC0		; gap following file name, skip
	AND #$3F		; mask top two bits
	LDY #$15		; set column count to 21
	STA (PNT),Y		; save file type character to current screen line
	LDA COLOR		; get current colour code
	STA (USER),Y		; save to colour RAM
	LDA #':'
LACFE	DEC FNLEN		; decrement file name length
LAD00	JSR LA1D2		; send character to saved output device
	JSR STOP		; scan stop key
	BEQ LAD1F		; if stop close logical file and restore previous output device
	JSR LA413		; get character from keyboard queue
	BEQ LACC0		; if no key pressed continue
LAD0D	JSR LA413		; get character from keyboard queue
	BEQ LAD0D		; loop until key pressed
	BNE LACC0		; branch always

; end of line reached

LAD14	JSR LA1D0		; output CR to saved output device
	LDY #$02		; consume 2 words, line link, line number
	STY CMP0		; set header/directory entry flag
	JMP LAC86		; continue with next entry

LAD1E	PLA			; pop saved byte
LAD1F	JSR B4CLRCHN		; close input and output channels
	LDA #$60		; set logical file to $60
	JSR B4CLOSE		; close a specified logical file
	JMP RESTOUT		; restore previous output device


;***********************************************************************************;
;
; get non-space character

GETNOTSP
	JSR LAD38		; get character
	CMP #' '		; compare with space
	BEQ GETNOTSP		; loop if space
	RTS


;***********************************************************************************;
;
; get character and compare with [CR], return Zb = 1 if so

GETCR	JSR CHRIN		; input character from channel
	CMP #$0D		; compare with [CR]
	RTS

LAD38	JSR GETCR		; get character and compare with [CR]
	BNE LAD6F		; return if not [CR]
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
	BCS MONPROMPT		; if >= [CR] display prompt
LAD41	DEX			; decrement table index
	BPL LAD5F		; if >= 0 continue
	JMP (MONERRV)		; handle syntax error (through vector)


;***********************************************************************************;
;
; handle syntax error

MONERR
	LDA #'?'
	JSR CHROUT		; output character to channel
	LDX #$F8		; set value for stack pointer
	TXS			; set stack then ..


;***********************************************************************************;
;
; display prompt

MONPROMPT
	LDA #$0D		; [CR]
	LDX #'.'
	JSR PRINTAX		; output prompt on new line
LAD56	JSR GETNOTSP		; get non-space character
	CMP #'.'
	BEQ LAD56		; loop while dot
	LDX #$07		; set number of table entries
LAD5F	CMP MONKEYS,X		; compare with table entry
	BNE LAD41
	EOR #'L'
	STA TEMPST+4		; save load/save flag
	LDA #$AE		; set command execute pointer high byte
	PHA			; push it
	LDA MONCMDS,X		; get command execute pointer low byte
	PHA			; push it
LAD6F	RTS			; execute command


;***********************************************************************************;
;
; print 16 bit value as hex

PRINTHEX16
	LDA STAL+1		; get high byte
	JSR PRINTHEX8		; output as hex digits
	LDA STAL		; get low byte
PRINTHEX8
	PHA			; save value
	LSR			; shift top nibble ..
	LSR
	LSR
	LSR			; .. into bottom nibble
	JSR LAD82		; output most significant nibble as hex digit
	PLA			; restore value
	AND #$0F		; mask off top nibble
LAD82	ORA #'0'
	CMP #'9'+1
	BCC LAD8A		; numeric, output it
	ADC #$06		; make hex digit
LAD8A	JMP CHROUT		; output character to channel


;***********************************************************************************;
;
; display .A, .X

PRINTAX
	JSR CHROUT		; output character to channel
	TXA			; copy second character to .A
	.BYTE $2C		; makes next line BIT $0DA9


;***********************************************************************************;
;
; output [CR]

PRINTCR
	LDA #$0D
	.BYTE $2C		; makes next line BIT $20A9


;***********************************************************************************;
;
; output space

PRINTSP
	LDA #' '
	JMP CHROUT		; output character to channel


;***********************************************************************************;
;
; get 8 bit hex value

GETHEX8
	LDA #$00		; clear .A
	STA TEMPST+3		; clear value
	JSR GETNOTSP		; get non-space character
	JSR LADA7		; convert first digit
	JSR LAD38		; get character
LADA7	CMP #'0'		; compare with "0"
	BCC MONERR		; if < do syntax error
	CMP #'9'+1
	AND #$0F		; convert to hex nibble
	BCC LADB3		; if a decimal digit skip
	ADC #'A'-'9'		; else adjust alphabetic
LADB3	ASL TEMPST+3		; shift value ..
	ASL TEMPST+3
	ASL TEMPST+3
	ASL TEMPST+3		; .. into high nibble
	ORA TEMPST+3		; OR in low nibble
	STA TEMPST+3		; save value
	SEC
LADC0	RTS


;***********************************************************************************;
;
; scan stop key, close input and output channels if pressed

CHKSTOP
	JSR STOP		; scan stop key
	BNE LADC0		; exit if no stop
	PHP			; save status
	JSR CLRCHN		; close input and output channels
	PLP			; restore status
	RTS

LADCC	LDX #TEMPST
	.BYTE $2C		; makes next line BIT $C1A2
LADCF	LDX #STAL
LADD1	JSR GETHEX8		; get 8 bit hex value
	STA $01,X		; set high byte
	JSR GETHEX8		; get 8 bit hex value
	STA $00,X		; set low byte
	RTS

LADDC	INC STAL		; increment low byte
	BNE LADE2		; if no rollover skip the high byte increment
	INC STAL+1		; increment high byte
LADE2	RTS

LADE3	LDA #<SAVEREG		; get saved registers low byte
	STA STAL		; set start address low byte
	LDA #>SAVEREG		; get saved registers high byte
	STA STAL+1		; set start address high byte
	LDA #$05		; number of registers
	RTS


;***********************************************************************************;
;
; print memory region as hex, STAL = start address, .A = length

PRINTMEM
	STA TEMPST+2		; set byte count
	LDY #$00		; clear index
LADF2	JSR PRINTSP		; print space
	LDA (STAL),Y		; get byte
	JSR PRINTHEX8		; print it as hex
	JSR LADDC		; increment start address pointer
	BEQ LAE04		; if high byte rolled over exit
	DEC TEMPST+2		; decrement byte count
	BNE LADF2		; loop until zero
	INY
LAE04	RTS

MONKEYS
	.BYTE "MRXG:;LS"

REGHEAD
	.BYTE ";.",$0d,"PS RY RX CA RS  CP   ",$0d


;***********************************************************************************;
;
; perform MONITOR dump memory

MONDUMPMEM
	JSR LADCF		; get start address
	LDX #MEMUSS
	JSR LADD1		; get end address

LAE2E	JSR CHKSTOP		; check for STOP key
	BEQ LAE98		; branch if [STOP]
	JSR PRINTCR		; print [CR]
	LDA #'.'
	LDX #':'
	JSR PRINTAX		; print memory dump prefix
	JSR PRINTHEX16
	LDA #$05		; set region length
	JSR PRINTMEM		; print memory region as hex
	BEQ LAE98		; if start address rolled over then exit
	LDA MEMUSS		; get end address low byte
	CMP STAL		; compare with start address low byte
	LDA MEMUSS+1		; get end address high byte
	SBC STAL+1		; subtract start address high byte
	BCS LAE2E		; loop until end reached
	BCC LAE98		; display prompt, branch always


;***********************************************************************************;
;
; BRK handler, enter into MONITOR

B4BREAK
	JSR CLRCHN		; close input and output channels
	PLA			; pull .Y
	STA SAVEREG+3		; set saved .Y
	PLA			; pull .X
	STA SAVEREG+2		; set saved .X
	PLA			; pull .A
	STA SAVEREG+1		; set saved .A
	PLA			; pull status
	STA SAVEREG		; set saved status
	PLA			; pull program counter low byte
	STA TEMPST		; set saved program counter low byte
	PLA			; pull program counter high byte
	STA TEMPST+1		; set saved program counter high byte
	TSX			; copy stack pointer
	STX SAVEREG+4		; set saved stack pointer
	LDA TEMPST		; get saved program counter low byte
	BNE LAE71		; branch if not zero
	DEC TEMPST+1		; decrement saved program counter high byte
LAE71	DEC TEMPST		; decrement saved program counter low byte
	JSR PRINTCR		; print [CR]
	LDA #'B'
	LDX #'*'
	JSR PRINTAX		; print break indicator


;***********************************************************************************;
;
; perform MONITOR dump registers

MONDUMPREG
	LDX #$18		; set length of heading
LAE7F	LDA REGHEAD,X		; get heading character
	JSR CHROUT		; print heading character
	DEX			; decrement index
	BPL LAE7F		; loop while more characters
	LDA TEMPST+1		; get saved program counter high byte
	JSR PRINTHEX8		; print as hex
	LDA TEMPST		; get saved program counter low byte
	JSR PRINTHEX8		; print as hex
	JSR LADE3		; set start address to saved registers
	JSR PRINTMEM		; print saved registers as hex
LAE98	JMP MONPROMPT		; display prompt


;***********************************************************************************;
;
; perform MONITOR set memory

MONSETMEM
	JSR LADCF		; get start address
	LDA #$05		; set byte count
	BNE LAEA8		; branch always


;***********************************************************************************;
;
; perform MONITOR set registers

MONSETREG
	JSR LADCC		; get source address
	JSR LADE3		; set destination address to saved registers
LAEA8	STA TEMPST+2		; set byte count
LAEAA	JSR GETHEX8		; get 8 bit hex value
	BCC LAEB7		; ??? dead code ???
	LDX #$00		; clear index
	STA (STAL,X)		; store value
	CMP (STAL,X)		; verify value
	BNE LAF11		; handle syntax error
LAEB7	JSR LADDC		; increment source address
	DEC TEMPST+2		; decrement byte count
	BNE LAEAA		; loop until done
	BEQ LAE98		; display prompt, branch always


;***********************************************************************************;
;
; perform MONITOR GO

MONGO
	JSR GETCR		; get character and compare with [CR]
	BEQ LAEC8		; branch if [CR]
	JSR LADCC		; read program counter
LAEC8	LDX SAVEREG+4		; get saved stack pointer
	TXS			; restore stack pointer
	LDA TEMPST+1		; get saved program counter high byte
	PHA			; store for return
	LDA TEMPST		; get saved program counter low byte
	PHA			; store for return
	LDA SAVEREG		; get saved status
	PHA			; store for return
	LDA SAVEREG+1		; restore .A
	LDX SAVEREG+2		; restore .X
	LDY SAVEREG+3		; restore .Y
	RTI			; restore status and return


;***********************************************************************************;
;
; perform MONITOR EXIT
	
MONEXIT
	JMP (WARMST)		; do BASIC break entry


;***********************************************************************************;
;
; perform MONITOR LOAD/SAVE

MONLDSV	LDY #$02
	STY FNADR+1		; set file name pointer high byte
	DEY
	STY FA			; set device number to 1 (tape)
	STY SA			; set secondary address to 1
	DEY			; clear .Y
	STY FNLEN		; set file name length to zero
	STY STATUS		; clear I/O status byte
	STY VERCK		; clear the load/verify flag
	LDA #$40
	STA FNADR		; set file name pointer low byte
LAEF2	JSR GETCR		; get character and compare with [CR]
	BEQ LAF14		; if [CR] do load
	CMP #' '		; compare with space
	BEQ LAEF2		; loop while space
	CMP #$22		; compare with quote character
	BNE LAF11		; if no file name do syntax error
LAEFF	JSR GETCR		; get character and compare with [CR]
	BEQ LAF14		; if [CR] do load
	CMP #$22		; compare with quote character
	BEQ LAF28		; if end quote check for other parameters
	STA (FNADR),Y		; save file name character
	INC FNLEN		; increment file name length
	INY			; increment index
	CPY #$13
	BNE LAEFF
LAF11	JMP MONERR		; handle syntax error

LAF14	LDA TEMPST+4		; get load/save flag
	BNE LAF11		; if save do syntax error
	LDA #$00		; flag load
	JSR LOAD		; load RAM from a device
	JSR CHKSTOP		; check for STOP key
	LDA STATUS		; read I/O status byte
	AND #$10		; mask 000x 0000, read error
	BNE LAF11		; if read error then exit
	BEQ LAF65		; display prompt, branch always

LAF28	JSR GETCR		; get character and compare with [CR]
	BEQ LAF14		; if [CR] do load
	CMP #','		; compare with ","
	BNE LAF11		; if not "," do syntax error
	JSR GETHEX8		; get 8 bit hex value
	AND #$0F		; mask low nibble
	BEQ LAF11		; if keyboard device do syntax error
	CMP #$03		; compare device number with screen
	BEQ LAF11		; if screen device do syntax error
	STA FA			; set device number
	JSR GETCR		; get character and compare with [CR]
	BEQ LAF14		; if [CR] do load
	CMP #','		; compare with ","
	BNE LAF11		; if not "," do syntax error
	JSR LADCF		; get start address
	JSR CHRIN		; input character from channel
	CMP #','		; compare with ","
	BNE LAF11		; if not "," do syntax error
	LDX #EAL
	JSR LADD1		; get end address
LAF56	JSR GETCR		; get character and compare with [CR]
	BNE LAF56		; loop if not [CR]
	LDA TEMPST+4		; get load/save flag
	BEQ LAF11		; if load do syntax error
	JSR PRINTCR		; print [CR]
	JSR $F682		; call ISAVE vector
LAF65	JMP MONPROMPT		; display prompt

MONCMDS
	.BYTE <MONDUMPMEM-1	; dump Memory command
	.BYTE <MONDUMPREG-1	; dump Registers command
	.BYTE <MONEXIT-1	; eXit command
	.BYTE <MONGO-1		; Go command
	.BYTE <MONSETMEM-1	; set memory command (:)
	.BYTE <MONSETREG-1	; set registers command (;)
	.BYTE <MONLDSV-1	; Load command
	.BYTE <MONLDSV-1	; Save command


;***********************************************************************************;
;
; perform MODEM

MODEM
	JSR $F118		; input from RS-232 buffer
	JSR RSNXTIN		; get byte from RS-232 buffer
	TAX			; copy byte to .X
	BEQ LAF99		; branch in no data
	AND #$7F		; clear top bit
	CMP #$08		; compare with ASCII BS
	BNE LAF81		; if not BS continue with input translation
	LDA #$14		; replace with PETSCII DEL
LAF81	CMP #"Z"+1		; compare with ASCII "Z" + 1
	BCC LAF89		; if upper case alpha numeric continue with input translation
	CMP #$5E		; compare with ASCII "^"
	BCC LAF96		; if "[", "\" or "]" go display it
LAF89	CMP #"@"		; compare with ASCII "@"
	BCC LAF96		; if not alphabetic go display it
	CMP #$60		; compare with ASCII "`"
	BCC LAF94		; if upper case alphabetic convert and display it
	AND #$5F		; convert to lower case
	.BYTE $2C		; makes next line BIT $8009
LAF94	ORA #$80		; convert to upper case
LAF96	JSR PRT2SCRN		; display received character to the screen
LAF99	LDA SHFLAG		; get keyboard shift/control/C= flag
	AND #$04		; mask [CTRL]
	BEQ LAFA9		; branch if not [CTRL]
LAFA0	JSR $F1F9		; call KERNAL routine to read keyboard
	BEQ LAFA0		; loop until key pressed
	AND #$1F		; mask to translate "A" to ctrl-A
	BPL LAFAE		; branch always
LAFA9	JSR $F1F9		; call KERNAL routine to read keyboard
	BEQ MODEM		; loop until key pressed
LAFAE	PHA			; save key
	LDA M51CDR		; get 6551 pseudo command register
	LSR			; ??? shift unused bit .. ???
	LSR			; .. into Cb
	PLA			; restore key
	BCC LAFBA		; branch if ???
	JSR PRT2SCRN		; display received character to the screen
LAFBA	BMI LAFD0		; if PETSCII graphic symbol go transmit it
	CMP #$14		; compare with PETSCII [DEL]
	BNE LAFC2		; if not DEL continue with output translation
	LDA #$08		; convert to ASCII BS
LAFC2	CMP #'Z'+1		; compare with PETSCII "Z" + 1
	BCC LAFCA		; if upper case alpha numeric continue with output translation
	CMP #$5E		; compare with PETSCII "↑"
	BCC LAFD0		; if "[", "£" or "]" go transmit it
LAFCA	CMP #$40		; compare with PETSCII "@"
	BCC LAFD0		; if not alphabetic go transmit it
	ORA #$60		; convert to lower case
LAFD0	AND #$7F		; mask top bit
	STA PTR1
	JSR RSOUTSAV		; send byte to RS-232 buffer
	JMP MODEM		; loop forever


;***********************************************************************************;
;
; display character to screen at current cursor position

PRT2SCRN
	PHA			; save character
	LDA #' '
	JSR PRINTAT		; print space to screen
	PLA			; restore character
	PHA			; save character
	CMP #$14		; compare with [DEL]
	BNE LAFE8		; branch if not [DEL]
	LDA #$9D		; [CRSR R]
LAFE8	JSR SCRNOUT		; output character and advance cursor
	LDA #$64		; PETSCII graphics char
	JSR PRINTAT		; print cursor to screen
	PLA			; restore saved character
	RTS


;***********************************************************************************;
;
; display character to screen at current cursor position

PRINTAT
	PHA			; save character
	JSR COLORSYN		; calculate pointer to colour RAM
	PLA			; restore character
	LDX COLOR		; get current colour code
	JMP PUTSCRN		; save character and colour to screen @ cursor

	.BYTE "DF "

;***********************************************************************************;
;***********************************************************************************;
;
; The almost completely commented VIC-1112 ROM disassembly. v1.3
; By Simon Rowe <srowe@mose.org.uk>.

; This is a bit correct assembly listing for the ROM in the VIC-20 IEEE-488 cartridge.


	; KERNAL vectors
NMINV	= $0318			; NMI vector
IOPEN	= $031A			; open a logical file
IGETIN	= $032A			; get character from keyboard queue
ILOAD	= $0330			; load
ISAVE	= $0332			; save

	; Standard VIA
VIA1IFR	= $911D			; VIA 1 IFR
					; bit	function		cleared by
					; ---	--------		----------
					;  7	interrupt status	clearing all enabled interrupts
					;  6	T1 interrupt		read T1C_l, write T1C_h
					;  5	T2 interrupt		read T2C_l, write T2C_h
					;  4	CB1 transition		read or write port B
					;  3	CB2 transition		read or write port B
					;  2	8 shifts done		read or write the shift register
					;  1	CA1 transition		read or write port A
					;  0	CA2 transition		read or write port A

VIA1IER	= $911E			; VIA 1 IER
					; bit	function
					; ---	--------
					;  7	enable/disable
					;  6	T1 interrupt
					;  5	T2 interrupt
					;  4	CB1 transition
					;  3	CB2 transition
					;  2	8 shifts done
					;  1	CA1 transition
					;  0	CA2 transition

	; cartridge VIA A (U4)
VIAAPB	= $9800			; VIA A DRB
					; bit	function
					; ---	--------
					;  7	ATN in
					;  6	NDAC in
					;  5	NRFD in
					;  4	DAV in
					;  3	EOI in
					;  2	NDAC out
					;  1	NRFD out
					;  0	DAV out

VIAADDRB = $9802		; VIA A DDRB
VIAADDRA = $9803		; VIA A DDRA
VIAAACR	= $980B			; VIA A ACR
					; bit	function
					; ---	--------

VIAAPCR	= $980C			; VIA A PCR
					; bit	function
					; ---	--------

VIAAIFR = $980D			; VIA A IFR
VIAAIER	= $980E			; VIA A IER
VIAAPA2	= $980F			; VIA A DRA, no handshake

	; cartridge VIA B (U5)
VIABPB	= $9810			; VIA B DRB
					; bit	function
					; ---	--------
					;  7	DIO8 in
					;  6	DIO7 in
					;  5	DIO6 in
					;  4	DIO5 in
					;  3	DIO4 in
					;  2	DIO3 in
					;  1	DIO2 in
					;  0	DIO1 in

VIABDDRB = $9812		; VIA B DDRB
VIABDDRA = $9813		; VIA B DDRA
VIABT1CH = $9815		; VIA B T1C_h
VIABACR	= $981B			; VIA B ACR
VIABPCR	= $981C			; VIA B PCR
					; bit	function
					; ---	--------
					; 765	CB2 EOI out
					;  4	CB1 SRQ in
					; 321	CA2 ATN out
					;  0	

VIABIFR	= $981D			; VIA B IFR
VIABIER	= $981E			; VIA B IER
VIABPA2	= $981F			; VIA B DRA, no handshake
					; bit	function
					; ---	--------
					;  7	DIO8 out
					;  6	DIO7 out
					;  5	DIO6 out
					;  4	DIO5 out
					;  3	DIO4 out
					;  2	DIO3 out
					;  1	DIO2 out
					;  0	DIO1 out

	; Unknown registers on IO3/ *** dead code ***
UNKWN1	= $9C00
UNKWN2	= $9C01

	; XROM entrypoints
XROMWARM = $A002		; autostart ROM break entry vector
	
	; BASIC entrypoints
READY	= $C474			; do warm start
PRTSTR	= $CB1E			; print null terminated string

	; KERNAL entrypoints
RSDVCERR = $F0B9		; do illegal device number
CHRINSR	= $F264			; get byte from serial bus
CHRINRS	= $F26F			; get byte from RS-232 device
SRCHING	= $F647			; display "SEARCHING..."
LDVRMSG	= $F66A			; display "LOADING" or "VERIFYING"
SAVING	= $F728			; display saving [file name]
FE_NTOUT = $F790		; display not output file
FE_ILDEV = $F796		; display illegal device number
CHKAUTO	= $FD3F			; scan for autostart ROM at $A000
FRESTOR	= $FD52			; restore default I/O vectors
RSNMI	= $FEDE			; RS-232 NMI routine
_RTI	= $FF56			; restore registers and exit interrupt

	
;***********************************************************************************;
;***********************************************************************************;
;
; IEEE-488 ROM start

	;* = $B000

	.WORD IEINIT
	.WORD IENMINV
	.BYTE "B0"
	NOP
	NOP
	NOP

IEINIT
	JSR INITIEVIA		; initialize cartridge I/O registers
	JSR INITVEC		; initialize KERNAL vectors to IEEE-488 functions
	LDA #<BANNER		; set banner pointer low byte
	LDY #>BANNER		; set banner pointer high byte
	JSR PRTSTR		; print banner
	LDX #$FB		; value for start stack
	TXS			; set stack pointer
	JMP READY		; return to BASIC

BANNER	.BYTE $0D,"-- IEEE-488 DRIVER --",$0D,$00


;***********************************************************************************;
;
; initialize KERNAL vectors

INITVEC
	LDX #$0D		; set index to vector table length
LB036	LDA VTABLE,X		; get vector
	STA IOPEN,X		; save vector
	DEX			; decrement index 
	BPL LB036		; loop until index < 0
	LDA #<IELOAD		; get IELOAD address low byte
	STA ILOAD		; save to ILOAD vector low byte
	LDA #>IELOAD		; get IELOAD address high byte
	STA ILOAD+1		; save to ILOAD vector high byte
	LDA #<IESAVE		; get IESAVE address low byte
	STA ISAVE		; save to ISAVE vector low byte
	LDA #>IESAVE		; get IESAVE address high byte
	STA ISAVE+1		; save to ISAVE vector high byte
	LDA #<IEGETIN		; get IEGETIN address low byte
	STA IGETIN		; save to IGETIN vector low byte
	LDA #>IEGETIN		; get IEGETIN address high byte
	STA IGETIN+1		; save to IGETIN vector high byte
	LDA #<IENMINV		; get IENMINV address low byte
	STA NMINV		; save to NMINV vector low byte
	LDA #>IENMINV		; get IENMINV address high byte
	STA NMINV+1		; save to NMINV vector high byte
	RTS

VTABLE	.WORD IEOPEN		; OPEN vector
	.WORD IECLOSE		; CLOSE vector
	.WORD IECHKIN		; CHKIN vector 
	.WORD IECHKOUT		; CHKOUT vector
	.WORD IECLRCHN		; CLRCHN vector
	.WORD IECHRIN		; CHRIN vector
	.WORD IECHROUT		; CHROUT vector


;***********************************************************************************;
;
; NMI handler

IENMINV
	SEI			; disable interrupts
	PHA			; save .A
	TXA			; copy .X
	PHA			; save .X
	TYA			; copy .Y
	PHA			; save .Y
	LDA VIA1IFR		; get VIA 1 IFR
	BMI LB084		; branch if interrupt raised
	JMP _RTI		; restore registers and exit interrupt
LB084	AND VIA1IER		; AND with VIA 1 IER
	TAX			; copy to .X
	AND #$02		; mask [RESTORE] key
	BNE LB08F		; branch if [RESTORE] key
	JMP RSNMI		; jump to KERNAL to handle RS-232 interrupt
LB08F	JSR CHKAUTO		; scan for autostart ROM at $A000
	BNE LB097		; branch if no autostart ROM
	JMP (XROMWARM)		; else do autostart ROM break entry
LB097	BIT VIA1PA1		; test VIA 1 DRA
	JSR FUDTIM		; increment the real time clock
	JSR STOP		; scan stop key
	BEQ IEBREAK		; branch if not [STOP]
	JMP _RTI		; restore registers and exit interrupt


;***********************************************************************************;
;
; BRK handler

IEBREAK
	JSR FRESTOR		; restore default I/O vectors
	JSR INITVIA		; initialize I/O registers
	JSR INITSK		; initialize hardware
	JSR INITIEVIA		; initialize cartridge I/O registers
	JSR INITVEC		; initialize KERNAL vectors to IEEE-488 functions
	JMP (WARMST)		; do BASIC break entry


;***********************************************************************************;
;
; open a channel for output

; Any logical file that has already been opened by the OPEN routine can be defined
; as an output channel by this routine the device on the channel must be an output
; output device or an error will occur and the routine will abort.

; If you are sending data to anywhere other than the screen this routine must be
; called before using the CHROUT routine. If you are sending data to the screen and
; no other output channels are open then the calls to this routine and to the OPEN
; routine are not needed.

; When used with a device on the IEEE bus this routine will automatically send the
; listen address specified by the OPEN routine and any secondary address.

; Possible errors are:
;
;	3 : file not open
;	5 : device not present
;	7 : file is not an output file

IECHKOUT
	JSR FNDFLNO		; find file
	BEQ LB0BF		; branch if file found
	JMP FE_NTOPN		; do file not open error and retirn
LB0BF	JSR SETFLCH		; set file details from table,X
	LDA FA			; get device number
	BNE LB0C9		; branch if device is not keyboard
	JMP FE_NTOUT		; do not output file error and return
LB0C9	CMP #$03		; compare device number with screen
	BEQ LB0D2		; if screen save output device number and exit
	BCS LB0D6		; branch if > screen, IEEE bus device
	JMP $F321		; jump to KERNAL to handle RS-232 and tape
LB0D2	STA DFLTO		; save output device number
	CLC			; flag ok
	RTS
LB0D6	JMP LB0E8		; skip dead code

	NOP			; *** dead code ***
	JMP $F332		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB0E8		; *** dead code ***
	JSR LB59D		; *** dead code ***
	STA DFLTO		; *** dead code ***
	CLC			; *** dead code ***
	RTS			; *** dead code ***

LB0E8	JSR LB40C		; command an IEEE device to LISTEN
	STA DFLTO		; save output device number
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; output a character to channel

; This routine will output a character to an already opened channel. Use the OPEN
; routine, OPEN, and the CHKOUT routine, to set up the output channel before calling
; this routine. If these calls are omitted, data will be sent to the default output
; device, device 3, the screen. The data byte to be output is loaded into the accumulator,
; and this routine is called. The data is then sent to the specified output device.
; The channel is left open after the call.

; NOTE: Care must be taken when using routine to send data to an IEEE device since
; data will be sent to all open output channels on the bus. Unless this is desired,
; all open output channels on the IEEE bus other than the actually intended
; destination channel must be closed by a call to the KERNAL close channel routine.

IECHROUT
	PHA			; save the character to send
	LDA DFLTO		; get output device number
	CMP #$03		; compare device number with screen
	BNE LB0FA		; if output device not screen continue
	PLA			; restore character to send
	JMP SCRNOUT		; output character and return
LB0FA	BCS LB0FF		; branch if IEEE bus device
	JMP $F28B		; jump to KERNEL to handle RS-232 and tape
LB0FF	JMP LB112		; skip dead code

	NOP			; *** dead code ***
	JMP $F287		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB112		; *** dead code ***
	PLA			; *** dead code ***
	STA BSOUR		; *** dead code ***
	JSR LB5AC		; *** dead code ***
	CLC			; *** dead code ***
	RTS			; *** dead code ***

LB112	PLA			; restore character to send
	STA BSOUR		; store byte to send
	JSR IECIOUT		; send byte on IEEE bus
	LDA BSOUR		; restore byte sent
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; open a channel for input

; Any logical file that has already been opened by the OPEN routine can be defined as
; an input channel by this routine. The device on the channel must be an input device
; or an error will occur and the routine will abort.

; If you are getting data from anywhere other than the keyboard, this routine must be
; called before using either the CHRIN routine or the GETIN routine. If you are
; getting data from the keyboard and no other input channels are open then the calls
; to this routine and to the OPEN routine are not needed.

; When used with a device on the IEEE bus this routine will automatically send the
; listen address specified by the OPEN routine and any secondary address.

; Possible errors are:
;
;	3 : file not open
;	5 : device not present
;	6 : file is not an input file

IECHKIN
	JSR FNDFLNO		; find file
	BEQ LB124		; branch if file opened
	JMP FE_NTOPN		; do file not open error and return
LB124	JSR SETFLCH		; set file details from table,X
	LDA FA			; get device number
	BEQ LB134		; if device was keyboard save device #, flag ok and exit
	CMP #$03		; compare device number with screen
	BEQ LB134		; if device was screen save device #, flag ok and exit
	BCS LB138		; branch if IEEE bus device
	JMP $F2DC		; jump to KERNAL to handle RS-232 and tape
LB134	STA DFLTN		; save input device number
	CLC			; flag ok
	RTS
LB138	JMP LB146		; skip dead code

	NOP			; *** dead code ***
	JMP $F2F0		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB146		; *** dead code ***
	JMP LB610		; *** dead code ***

LB146	JSR LB34D		; command an IEEE device to TALK
	STA DFLTN		; save input device number
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; input character from channel

; This routine will get a byte of data from the channel already set up as the input
; channel by the CHKIN routine.

; If CHKIN has not been used to define another input channel the data is expected to be
; from the keyboard. the data byte is returned in the accumulator. The channel remains
; open after the call.

IECHRIN
	LDA DFLTN		; get input device number
	BNE LB154		; if it's not the keyboard continue
	JMP $F212		; jump to KERNAL to handle keyboard input
LB154	CMP #$03		; compare device number with screen
	BNE LB15B		; if it's not the screen continue
	JMP $F221		; jump to KERNAL to handle screen input
LB15B	BCS LB160		; branch if IEEE bus
	JMP $F22C		; jump to KERNAL to handle RS-232 and tape
LB160	JMP LB16E		; skip dead code

	NOP			; *** dead code ***
	JMP CHRINSR		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB16E		; *** dead code ***
	JMP CHRINRS		; *** dead code ***

LB16E	LDA STATUS		; get I/O status byte
	BEQ LB176		; if no errors flagged go input byte and return
	LDA #$0D		; else return [EOL]
	CLC			; flag ok
	RTS

LB176	JSR IEACPTR		; input a byte from the IEEE bus
	LDA BSOUR		; get received byte
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; close input and output channels

; This routine is called to clear all open channels and restore the I/O channels to
; their original default values. It is usually called after opening other I/O
; channels and using them for input/output operations. The default input device is
; 0, the keyboard. The default output device is 3, the screen.

; If one of the channels to be closed is to the IEEE bus, an UNTALK signal is sent
; first to clear the input channel or an UNLISTEN is sent to clear the output channel.
; By not calling this routine and leaving listener(s) active on the IEEE bus,
; several devices can receive the same data from the VIC at the same time. One way to
; take advantage of this would be to command the printer to LISTEN and the disk to
; TALK. This would allow direct printing of a disk file.

IECLRCHN
	LDX #$03		; set .X to screen
	CPX DFLTO		; compare output device number with screen
	BCS LB198		; branch if >= screen
	NOP			; else was IEEE bus
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	JSR IEUNLSN		; command the IEEE bus to UNLISTEN
	JMP LB198		; skip dead code

	JSR $FFAE		; *** dead code ***

LB198	LDX #$03		; set .X to screen
	CPX DFLTN		; compare input device number with screen
	BCS LB1B3		; branch if >= screen
	NOP			; else was IEEE bus
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	JSR IEUNTLK		; send UNTALK command
	JMP LB1B3		; set default devices

	JSR $EEF9		; *** dead code ***

LB1B3	LDX #$03		; set .X to screen
	STX DFLTO		; set output device number to screen
	LDA #$00		; set .X to keyboard
	STA DFLTN		; set input device number to keyboard
	RTS


;***********************************************************************************;
;
; close a specified logical file

; This routine is used to close a logical file after all I/O operations have been
; completed on that file. This routine is called after the accumulator is loaded
; with the logical file number to be closed, the same number used when the file was
; opened using the OPEN routine.

IECLOSE
	JSR $F3D4		; find file .A
	BEQ LB1C3		; if the file is found go close it
	CLC			; else the file was closed so just flag ok
	RTS

;  found the file so close it

LB1C3	JSR SETFLCH		; set file details from table,X
	TXA			; copy file index to .A
	PHA			; save file index
	LDA FA			; get device number
	BEQ LB1D5		; if $00, keyboard, restore index and close file
	CMP #$03		; compare device number with screen
	BEQ LB1D5		; if screen restore index and close file
	BCS LB1D8		; if > screen go do IEEE bus device close
	JMP $F360		; else jump to KERNAL to handle RS-232 and tape
LB1D5	JMP $F3B1		; close file index .X
LB1D8	JMP LB1E6		; skip dead code

	NOP			; *** dead code ***
	JMP $F3AE		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB1E6		; *** dead code ***
	JMP LB5F6		; *** dead code ***

LB1E6	JSR LB3BB		; close IEEE bus device
	JMP $F3B1		; close file index .X


;***********************************************************************************;
;
; open a logical file

; This routine is used to open a logical file. Once the logical file is set up it
; can be used for input/output operations. Most of the I/O KERNAL routines call on
; this routine to create the logical files to operate on. No arguments need to be
; set up to use this routine, but both the SETLFS and SETNAM KERNAL routines must
; be called before using this routine.

IEOPEN
	LDX LA			; get logical file
	BNE LB1F3		; branch if there is a file
	JMP FE_NTINP		; else do not input file error and return
LB1F3	JSR FNDFLNO		; find file
	BNE LB1FB		; branch if file not found
	JMP FE_ALOPN		; else do file already open error and return
LB1FB	LDX LDTND		; get open file count
	CPX #$0A		; compare with max
	BCC LB204		; branch if less
	JMP FE_2MNYF		; else do too many files error and return
LB204	INC LDTND		; increment open file count
	LDA LA			; get logical file
	STA LAT,X		; save to logical file table
	LDA SA			; get secondary address
	ORA #$60		; OR with the OPEN CHANNEL command
	STA SA			; set secondary address
	STA SAT,X		; save to secondary address table
	LDA FA			; get device number
	STA FAT,X		; save to device number table
	BEQ LB237		; do ok exit if keyboard
	CMP #$03		; compare device number with screen
	BEQ LB237		; do ok exit if screen
	BCC LB234		; branch if < screen, tape or RS-232
	JMP LB22F		; skip dead code

	NOP			; *** dead code ***
	JMP $F43F		; *** dead code ***
	CMP #$09		; *** dead code ***
	BNE LB22F		; *** dead code ***
	JMP LB5E3		; *** dead code ***

LB22F	JSR IESERNAME		; send secondary address and filename
	CLC			; flag ok
	RTS

LB234	JMP $F444		; jump to KERNAL to handle RS-232 and tape
LB237	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; initialize cartridge I/O registers

INITIEVIA
	LDA #$00		; T1 & T2 1 shot, shift disabled, PA & PB latch disabled
	STA VIABACR		; set VIA B ACR
	STA VIAAACR		; set VIA A ACR

	STA VIABIFR		; clear pending VIA B interrupts
	STA VIAAIFR		; clear pending VIA A interrupts
	LDA #$FF		; CB2 high, CB1 +ve edge, CA2 high, CA1 +ve edge

; The above code is wrong, to clear all pending interrupts all 1s should be written
;
;	LDA #$FF
;	STA VIABIFR		; clear pending VIA B interrupts
;	STA VIAAIFR		; clear pending VIA A interrupts

	STA VIABPCR		; set VIA B PCR
	STA VIAAPCR		; set VIA A PCR (unused)

	LDA #$80		; enable interrupts, no bits affected
	STA VIABIER		; modify VIA B interrupts
	STA VIAAIER		; modify VIA A interrupts

; The above code is wrong, to disable all interrupts b7 should be 0 and b6-0 should be 1s
;
;	LDA #7F			; disable interrupts, all bits affected
;	STA VIABIER		; modify VIA B interrupts
;	STA VIAAIER		; modify VIA A interrupts

	LDA #$07		; IIIIIOOO, ATN in, NDAC in, NRFD in, DAV in,
				; EOI in, NDAC out, NRFD out, DAV out
	STA VIAADDRB		; set VIA A DDRB
	STA VIAAPB		; set VIA A DRB outputs high
	LDA #$FF		; all outputs
	STA VIABDDRA		; set VIA B DDRA, DIO1-8
	STA VIAADDRA		; set VIA A DDRA (unused)
	STA VIABPA2		; set VIA B DRA, all ones
	STA VIAAPA2		; set VIA A DRA (unused)
	LDA #$00		; all inputs
	STA VIABDDRB		; set VIA B DDRB, DIO1-8
	RTS

	LDA #$FD		; *** dead code ***
	STA BSOUR		; *** dead code ***
	LDA BSOUR		; *** dead code ***
	STA UNKWN2		; *** dead code ***
LB27C	LDA #$40		; *** dead code ***
	STA UNKWN2		; *** dead code ***
	LDA BSOUR		; *** dead code ***
	STA UNKWN2		; *** dead code ***
	LDA #$12		; *** dead code ***
	STA UNKWN2		; *** dead code ***
	RTS			; *** dead code ***

; OR into I/O status byte

IEORIOST
	ORA STATUS		; OR with I/O status byte
	STA STATUS		; save I/O status byte
	RTS

LB291	LDA VIAAPB		; get VIA A DRB
	ORA #$02		; set NRFD out high
	STA VIAAPB		; set VIA A DRB
	LDA #$FF		; set for 65,280 cycles
	STA VIABT1CH		; set VIA B T1C_h
LB29E	LDA VIAAPB		; get VIA A DRB
	AND #$10		; mask DAV in
	BEQ LB2B1		; branch if DAV low
	LDA VIABIFR		; get VIA B IFR
	AND #$40		; mask T1 interrupt
	BEQ LB29E		; loop while timer not expired
	LDA #$02		; set time out read
	STA STATUS		; set I/O status byte
	RTS

; receive byte from IEEE bus

LB2B1	LDA VIAAPB		; get VIA A DRB
	AND #$08		; mask EOI in
	BNE LB2BE		; branch if not EOI
	LDA #$40		; set EOI bit
	JSR IEORIOST		; OR into I/O status byte
	NOP
LB2BE	LDA VIABPB		; get VIA B DRB
	EOR #$FF		; invert it
	STA BSOUR		; save byte received
	LDA VIAAPB		; get VIA A DRB
	AND #$FD		; mask NRFD out
	STA VIAAPB		; store VIA A DRB
	LDA VIAAPB		; get VIA A DRB
	ORA #$04		; set NDAC out
	STA VIAAPB		; store VIA A DRB
LB2D5	LDA VIAAPB		; get VIA A DRB
	AND #$10		; mask DAV in
	BEQ LB2D5		; loop if DAV low
	LDA VIAAPB		; get VIA A DRB
	AND #$F9		; mask NRFD and NDAC out
	STA VIAAPB		; store VIA A DRB
	RTS

; output byte on IEEE bus

LB2E5	LDA BSOUR		; get byte to output
	EOR #$FF		; invert it
	STA VIABPA2		; set VIA B DRA
LB2EC	LDA VIAAPB		; get VIA A DRB
	AND #$20		; mask NRFD in
	BEQ LB2EC		; branch if NRFD low
	LDA VIAAPB		; get VIA A DRB
	AND #$FE		; clear DAV out
	STA VIAAPB		; store VIA A DRB
	LDA #$FF		; set for 65,280 cycles
	STA VIABT1CH		; set VIA B T1C_h
LB300	LDA VIAAPB		; get VIA A DRB
	AND #$40		; mask NDAC in
	BNE LB313		; branch if NDAC high
	LDA VIABIFR		; get VIA B IFR
	AND #$40		; mask T1 interrupt
	BEQ LB300		; loop while timer not expired
	LDA #$01		; set time out write
	STA STATUS		; save I/O status byte
	RTS

LB313	LDA VIAAPB		; get VIA B IFR
	ORA #$01		; set DAV out high
	STA VIAAPB		; store VIA A DRB
	LDA #$FF		; set data to $00
	STA VIABPA2		; set VIA B DRA
	CLC			; flag ok
	RTS

; output final byte on IEEE bus

LB322	LDA VIABPCR		; get VIA B PCR
	AND #$DF		; set CB2 low, EOI low
	STA VIABPCR		; save VIA B PCR
	JSR LB2E5		; output a byte to the IEEE bus
	LDA VIABPCR		; get VIA B PCR
	ORA #$20		; set CB2 high, EOI high
	STA VIABPCR		; save VIA B PCR
	RTS


;***********************************************************************************;
;
; command an IEEE bus device to TALK

; To use this routine the accumulator must first be loaded with a device number
; between 4 and 30. When called this routine converts this device number to a talk
; address. Then this data is transmitted as a command on the IEEE bus.

IETALK
	ORA #$40		; OR with the TALK command
	.BYTE $2C		; makes next line BIT $2009


;***********************************************************************************;
;
; command devices on the IEEE bus to LISTEN

; This routine will command a device on the IEEE bus to receive data. The
; accumulator must be loaded with a device number between 4 and 31 before calling
; this routine. LISTEN convert this to a listen address then transmit this data as
; a command on the IEEE bus. The specified device will then go into listen mode
; and be ready to accept information.

IELISTEN
	ORA #$20		; OR with the LISTEN command
LB33B	STA BSOUR		; store byte to send
	JSR LB381		; set ATN low and check device is present
	JSR LB2E5		; output a byte to the IEEE bus
	RTS


;***********************************************************************************;
;
; send IEEE secondary command

; ATN will have been set low by a previous LISTEN or TALK command, return it to high
; on completion.

IESECOND
	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
	JSR LB398		; set ATN high
	RTS

LB34D	TAX			; copy device number to .X
	JSR IETALK		; command an IEEE bus device to TALK
	LDA SA			; get secondary address
	BPL LB35B		; branch if address to send
	NOP
	NOP
	NOP
	JMP LB366		; check device is present
LB35B	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
LB366	LDA VIAAPB		; get VIA B DRB
	AND #$F9		; mask NRFD and NDAC out
	STA VIAAPB		; save VIA B DRB
	LDY #$10		; set delay loop interations
LB370	DEY			; decrement loop count
	BNE LB370		; branch while non-zero
	JSR LB398		; set ATN high
	TXA			; copy device back to .A
	BIT STATUS		; test I/O status byte
	BPL LB380		; if device present save device number and exit
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
	JMP FE_DVNTP		; do device not present error and return
LB380	RTS

LB381	LDA VIABPCR		; get VIA B PCR
	AND #$FD		; set CA2 low, ATN low
	STA VIABPCR		; save VIA B PCR
	LDA VIAAPB		; get VIA A DRB
	AND #$60		; mask NRFD and NDAC in
	CMP #$60		; test both NRFD and NDAC high
	BNE LB397		; branch if not both high
	LDA #$80		; set device not present bit
	JSR IEORIOST		; OR into I/O status byte
LB397	RTS

LB398	LDA VIABPCR		; get VIA B PCR
	ORA #$02		; set CA2 high, ATN high
	STA VIABPCR		; save VIA B PCR
	RTS


;***********************************************************************************;
;
; command the IEEE bus to UNLISTEN

; This routine commands all devices on the IEEE bus to stop receiving data from
; the computer. Calling this routine results in an UNLISTEN command being transmitted
; on the IEEE bus. Only devices previously commanded to listen will be affected.

; This routine is normally used after the computer is finished sending data to
; external devices. Sending the UNLISTEN will command the listening devices to get
; off the IEEE bus so it can be used for other purposes.

IEUNLSN
	LDA #$3F		; set the UNLISTEN command
	JSR LB33B		; send command
	JSR LB398		; set ATN high
	RTS


;***********************************************************************************;
;
; command the IEEE bus to UNTALK

; This routine will transmit an UNTALK command on the serial bus. All devices
; previously set to TALK will stop sending data when this command is received.

IEUNTLK
	LDA VIAAPB		; get VIA A DRB
	ORA #$06		; set NDAC out and NRFD out high
	STA VIAAPB		; set VIA A DRB
	LDA #$5F		; set the UNTALK command
	JSR LB33B		; send command
	JSR LB398		; set ATN high
	RTS

; close the IEEE bus device

LB3BB	BIT SA			; test the secondary address
	BMI LB3D0		; if already closed just exit
	LDA FA			; get device number
	JSR IELISTEN		; command devices on the IEEE bus to LISTEN
	LDA SA			; get secondary address
	AND #$EF		; mask the channel number
	ORA #$E0		; OR with the CLOSE command
	JSR IESECOND		; send IEEE secondary command
	JSR IEUNLSN		; command the IEEE bus to UNLISTEN
LB3D0	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; send secondary address and filename

IESERNAME
	LDA SA			; get secondary address
	BMI LB40A		; ok exit if no address
	LDY FNLEN		; get file name length
	BEQ LB40A		; ok exit if null
	LDA FA			; get device number
	JSR IELISTEN		; command devices on the IEEE bus to LISTEN
	LDA SA			; get the secondary address
	ORA #$F0		; OR with the OPEN command
	JSR IESECOND		; send IEEE secondary command
	LDA STATUS		; get I/O status byte
	BPL LB3EF		; branch if device present
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
	JMP FE_DVNTP		; do device not present error and return

LB3EF	LDA FNLEN		; get file name length
	BEQ LB407		; branch if null name
	LDY #$00		; clear index
LB3F5	LDA (FNADR),Y		; get file name byte
	STA BSOUR		; store byte to send
	INY			; increment index
	CPY FNLEN		; compare with file name length
	BEQ LB404		; branch if loop done
	JSR LB2E5		; output a byte to the IEEE bus
	JMP LB3F5		; continue processing file name
LB404	JSR LB322		; output a byte with EOI to the IEEE bus
LB407	JSR IEUNLSN		; command the IEEE bus to UNLISTEN
LB40A	CLC			; flag ok
	RTS

LB40C	TAX			; copy device number
	JSR IELISTEN		; command devices on the IEEE bus to LISTEN
	LDA SA			; get secondary address
	BPL LB41A		; branch if address to send
	JSR LB398		; set ATN high
	JMP LB41D		; no address to send
LB41A	JSR IESECOND		; send IEEE secondary command
LB41D	TXA			; copy device number back to .A
	BIT STATUS		; test I/O status byte
	BPL LB427		; if device present save output device number and exit
	PLA			; else dump calling address low byte
	PLA			; dump calling address high byte
	JMP FE_DVNTP		; do device not present error and return
LB427	RTS


;************************************************************************************
;
; output a byte to the IEEE bus

; This routine is used to send information to devices on the IEEE bus. A call to
; this routine will put a data byte onto the IEEE bus using full handshaking.
; Before this routine is called the LISTEN routine must be used to command a device
; on the IEEE bus to get ready to receive data.

; A byte is loaded from BSOURCE and output as data on the IEEE bus. A device
; must be listening or the status word will return a timeout. EOI is asserted at
; the end of every line, for LFNs greater than 127 with Line Feeds, otherwise with
; Carriage Returns.

IECIOUT
	LDA LA			; get logical file
	BMI LB434		; branch if top bit set, BASIC sends [LF] after [CR]
	LDA BSOUR		; get byte to output
	CMP #$0D		; compare with [CR]
	BEQ LB445		; output a byte with EOI and return
	BNE LB44A		; branch always
LB434	LDA BSOUR		; get byte to output
	CMP #$0A		; compare with [LF]
	BEQ LB445		; output a byte with EOI and return
	BNE LB44A		; branch always

	BIT $94			; *** dead code ***
	BMI LB445		; *** dead code ***
	SEC			; *** dead code ***
	ROR $94			; *** dead code ***
	BNE LB44A		; *** dead code ***

LB445	JSR LB322		; output a byte with EOI to the IEEE bus
	CLC			; flag ok
	RTS

LB44A	JSR LB2E5		; output a byte to the IEEE bus
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; input a byte from the IEEE bus

; This routine reads a byte of data from the IEEE bus using full handshaking. The
; data is returned in the accumulator. Before using this routine the TALK routine
; must have been called first to command the device on the IEEE bus to send data on
; the bus. If the input device needs a secondary command it must be sent before
; calling this routine.

; Errors are returned in the status word which can be read by calling the READST
; routine.

IEACPTR
	JSR LB291		; input a byte from the IEEE bus
	CLC			; flag ok
	RTS


;***********************************************************************************;
;
; load RAM from a device

; This routine will load data bytes from any input device directly into the memory
; of the computer. It can also be used for a verify operation comparing data from a
; device with the data already in memory, leaving the data stored in RAM unchanged.

; The accumulator must be set to 0 for a load operation or 1 for a verify. If the
; input device was OPENed with a secondary address of 0 the header information from
; device will be ignored. In this case .X.Y must contain the starting address for the
; load. If the device was addressed with a secondary address of 1 or 2 the data will
; load into memory starting at the location specified by the header. This routine
; returns the address of the highest RAM location which was loaded.

; Before this routine can be called, the SETLFS and SETNAM routines must be called.

IELOAD
	STA VERCK		; save load/verify flag
	LDA #$00		; clear .A
	STA STATUS		; clear I/O status byte
	LDA FA			; get device number
	BNE LB461		; branch if not keyboard
LB45E	JMP FE_ILDEV		; do illegal device number and return
LB461	CMP #$03		; compare device number with screen
	BEQ LB45E		; if screen go do illegal device number and return
	BCC LB46E		; branch if less than screen
	JMP LB471		; jump to handle IEEE bus

	NOP			; *** dead code ***
	JMP $F55C		; *** dead code ***

LB46E	JMP $F5CA		; jump to KERNAL to handle tape

LB471	LDY FNLEN		; get file name length
	BNE LB478		; branch if not null name
	JMP FE_MISFN		; else do missing file name error and return
LB478	JSR LB502		; get secondary address and print "Searching..."
	LDA #$60		; set secondary address to $00
	STA SA			; save secondary address
	JSR IESERNAME		; send secondary address and filename
	LDA FA			; get device number
	JSR IETALK		; command an IEEE bus device to TALK
	LDA SA			; get secondary address
	JSR IESECOND		; send IEEE secondary command
	LDA VIAAPB		; get VIA A DRB
	AND #$F9		; mask NRFD and NDAC out
	STA VIAAPB		; store VIA A DRB
	JSR IEACPTR		; input a byte from the IEEE bus
	LDA BSOUR		; get received byte
	STA EAL			; save program start address low byte
	LDA STATUS		; get I/O status byte
	LSR			; shift time out read ..
	LSR			; .. into carry bit
	BCS LB4F4		; if timed out go do file not found error and return
	JSR IEACPTR		; input a byte from the IEEE bus
	LDA BSOUR		; get received byte
	STA EAL+1		; save program start address high byte
	JSR LB507		; set LOAD address if secondary address = 0
LB4AB	LDA #$FD		; mask xxxx xx0x, clear time out read bit
	AND STATUS		; mask I/O status byte
	STA STATUS		; set I/O status byte
	JSR STOP		; scan stop key, return Zb = 1 = [STOP]
	BNE LB4C1		; branch if not [STOP]
	LDA VIAAPB		; get VIA A DRB
	ORA #$06		; set NDAC out and NRFD out high
	STA VIAAPB		; set VIA A DRB

	JMP $F6CB		; close the serial bus device and flag stop

; The above code is wrong, it should close the IEEE bus and exit
;
;	JMP LB572		; close the IEEE bus device and flag stop

LB4C1	JSR IEACPTR		; input a byte from the IEEE bus
	LDA BSOUR		; get received byte
	PHA
	PLA
	TAX			; copy byte
	LDA STATUS		; get I/O status byte
	LSR			; shift time out read ..
	LSR			; .. into carry bit
	BCS LB4AB		; if timed out clear I/O status and retry
	TXA			; copy received byte back
	LDY VERCK		; get load/verify flag
	BEQ LB4E0		; branch if load
	LDY #$00		; clear index
	CMP (EAL),Y		; compare byte with previously loaded byte
	BEQ LB4E2		; branch if match
	LDA #$10		; set read error bit
	JSR ORIOST		; OR into I/O status byte
	.BYTE $2C		; makes next line BIT $AE91
LB4E0	STA (EAL),Y		; save byte to memory
LB4E2	INC EAL			; increment save pointer low byte
	BNE LB4E8		; if no rollover skip the high byte increment
	INC EAL+1		; else increment save pointer high byte
LB4E8	BIT STATUS		; test I/O status byte
	BVC LB4AB		; loop if not end of file
	JSR IEUNTLK		; send UNTALK command
	JSR LB3BB		; close IEEE bus device
	BCC LB4FF		; branch if ok
LB4F4	LDA VIAAPB		; get VIA A DRB
	ORA #$06		; set NDAC out and NRFD out high
	STA VIAAPB		; set VIA A DRB
	JMP FE_NTFND		; do file not found error and return

LB4FF	JMP $F641		; flag ok and exit


;***********************************************************************************;
;
; get secondary address and print "Searching..."

LB502	LDX SA			; get secondary address
	JMP SRCHING		; print "Searching..." and return


;***********************************************************************************;
;
; set LOAD address if secondary address = 0

LB507	TXA			; copy secondary address
	BNE LB512		; load location not set in LOAD call, so
				; continue with load
	LDA MEMUSS		; get load start address low byte
	STA EAL			; save program start address low byte
	LDA MEMUSS+1		; get load start address high byte
	STA EAL+1		; save program start address high byte
LB512	JMP LDVRMSG		; display "LOADING" or "VERIFYING" and return


;***********************************************************************************;
;
; save RAM to device, .A = index to start address, .X.Y = end address low/high

; This routine saves a section of memory. Memory is saved from an indirect address
; on page 0 specified by A, to the address stored in .X.Y, to a logical file. The
; SETLFS and SETNAM routines must be used before calling this routine. However, a
; file name is not required to SAVE to device 1, the cassette. Any attempt to save to
; other devices without using a file name results in an error.

IESAVE
	LDA FA			; get device number
	BNE LB51C		; branch if not keyboard
LB519	JMP FE_ILDEV		; else do illegal device number and return
LB51C	CMP #$03		; compare device number with screen
	BEQ LB519		; if screen do illegal device number and return
	BCC LB529		; branch if < screen
	JMP LB533		; jump to handle IEEE bus

	NOP			; *** dead code ***
	JMP $F692		; *** dead code ***

LB529	CMP #$02		; compare device with RS-232 device
	BNE LB530		; branch if not RS-232 device
	JMP RSDVCERR		; else do illegal device number and return

LB530	JMP $F6F8		; jump to KERNAL to handle tape

LB533	LDA #$61		; set secondary address to $01
	STA SA			; save secondary address
	LDY FNLEN		; get file name length
	BNE LB53E		; branch if filename not null
	JMP FE_MISFN		; else do missing file name error and return
LB53E	JSR IESERNAME		; send secondary address and filename
	JSR SAVING		; print saving [file name]
	LDA FA			; get device number
	JSR IELISTEN		; command devices on the IEEE bus to LISTEN
	LDA SA			; get secondary address
	JSR IESECOND		; send IEEE secondary command
	LDY #$00		; clear index
	JSR RD300		; copy I/O start address to buffer address
	LDA SAL			; get buffer address low byte
	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
	LDA SAL+1		; get buffer address high byte
	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
LB561	JSR VPRTY		; check read/write pointer, return Cb = 1 if pointer >= end
	BCS LB57E		; go do UNLISTEN if at end
	LDA (SAL),Y		; get byte from buffer
	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
	JSR STOP		; scan stop key
	BNE LB579		; if stop not pressed go increment pointer and loop for next
LB572	JSR LB3BB		; close IEEE bus device
	LDA #$00		; terminated by STOP key
	SEC			; flag stop
	RTS

LB579	JSR WRT62		; increment read/write pointer
	BNE LB561		; loop, branch always

LB57E	JSR IEUNLSN		; command the IEEE bus to UNLISTEN
	JSR LB3BB		; close IEEE bus device
	RTS


;***********************************************************************************;
;
; get a character from the input device

; In practice this routine operates identically to the CHRIN routine for all devices
; except for the keyboard. If the keyboard is the current input device this routine
; will get one character from the keyboard buffer. It depends on the IRQ routine to
; read the keyboard and put characters into the buffer.

IEGETIN
	LDA DFLTN		; get input device number
	BNE LB58C		; branch if not keyboard
	JMP $F1F9		; jump to KERNAL to handle keyboard

LB58C	CMP #$02		; compare device with RS-232 device
	BNE LB593		; branch if not RS-232 device
	JMP $F205		; jump to KERNAL to handle RS-232

LB593	JMP LB59A		; skip dead code

	NOP			; *** dead code ***
	JMP $F205		; *** dead code ***

LB59A	JMP LB154		; jump to IECHRIN


;***********************************************************************************;
;
; dead code

LB59D	TAX
	LDA #$33
	STA UNKWN2
LB5A3	LDA UNKWN2
	AND #$80
	BEQ LB5A3
	TXA
	RTS
LB5AC	SEI
	LDA BSOUR
	CMP #$0D
	BNE LB5CC
	JSR LB63B
	LDA BSOUR
	STA UNKWN1
LB5BB	LDA UNKWN2
	AND #$04
	BEQ LB5BB
	LDA #$12
	STA UNKWN2
	LDA BSOUR
	CLI
	CLC
	RTS
LB5CC	JSR LB63B
	LDA BSOUR
	STA UNKWN1
	CLI
	CLC
	RTS
	LDA #$32
	STA UNKWN2
	RTS
	LDA #$32
	STA UNKWN2
	RTS
LB5E3	LDY #$00
	STY $0297
	CPY FNLEN
	BEQ LB5F3
	LDA (FNADR),Y
	STA BSOUR
	JSR LB27C
LB5F3	JMP $F51B
LB5F6	PLA
	JSR $F3B2
	LDA #$FD
	STA BSOUR
	JSR LB27C
	SEI
	LDA #<NMI2
	STA NMINV
	LDA #>NMI2
	STA NMINV+1
	CLI
	JMP $F377
LB610	STA DFLTN
	SEI
	LDA #<LB625
	STA NMINV
	LDA #>LB625
	STA NMINV+1
	CLI
	LDA #$36
	STA UNKWN2
	CLC
	RTS
LB625	PHA
	TXA
	PHA
	TYA
	PHA
	LDA UNKWN1
	LDY $029B
	STA ($F7),Y
	INC $029B
	PLA
	TAY
	PLA
	TAX
	PLA
	RTI
LB63B	LDA UNKWN2
	AND #$01
	BEQ LB63B
	RTS
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA
	LSR $94
	JMP $FF8A
	.BYTE $AA,$AA,$AA,$AA,$AA
	BIT $94
	BPL LB7B0
	SEC
	ROR $A3
	JSR $EE49
	LSR $94
	LSR $A3
LB7B0	JSR $B239
	JMP $B034
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA,$AA,$AA,$AA,$AA,$AA,$AA
	.BYTE $AA,$AA

IETKSA
	LDA SA			; get secondary address
	STA BSOUR		; store byte to send
	JSR LB2E5		; output a byte to the IEEE bus
	LDA VIAAPB		; get VIA A DRB
	AND #$F9		; mask NRFD and NDAC out
	STA VIAAPB		; store VIA A DRB
	JSR LB398		; set ATN high
	RTS

PATCH_NTFND
	JSR QUERYBUS		; query bus to use
	BCC SRNTFND		; branch if serial
	LDA VIAAPB		; get VIA A DRB
	ORA #$06		; set NDAC out and NRFD out high
	STA VIAAPB		; set VIA A DRB
SRNTFND	JMP FE_NTFND		; do file not found error and return
